#~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*#
# train.tcl                                                                #
#                                                                          #
# Description:                                                             #
#                                                                          #
#     This Tcl/Tk script is used to interactively generate handwriting     #
#     samples for use in character recognition, and also for testing the   #
#     recognition algorithm.                                               #
#                                                                          #
# Changes:                                                                 #
#                                                                          #
#     09/20/99  arvo    Split into smaller Tcl files.                      #
#     09/01/99  arvo    Added feature statistics visualizer.               #
#     08/18/99  arvo    Completely rewritten.                              #
#     08/15/99  arvo    Added profiling features to interface.             #
#     12/20/96  guo     Initial coding by Yuanshan Guo.                    #
#                                                                          #
#--------------------------------------------------------------------------#
# Copyright (C) 1999, James Arvo                                           #
#                                                                          #
# This program is free software; you can redistribute it and/or modify it  #
# under the terms of the GNU General Public License as published by the    #
# Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      #
#                                                                          #
# This program is distributed in the hope that it will be useful, but      #
# WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    #
# any particular purpose.  See the GNU General Public License for more     #
# details.                                                                 #
#                                                                          #
#~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*#

# Include loads a Tcl file from the same directory in which this Tcl script
# lives.  It constructs the full pathname to the present directory by 
# joining the current working directory with the path qualifier used in the
# original program invocation.  The resulting path is then used as the path
# qualifier for the given file name.
#
proc Include { filename } {
  global argv0
  set dir [file join [pwd] [file dirname $argv0]]
  if [catch "uplevel source [file join $dir $filename.tcl]"] then {
    puts "STARTUP ERROR: Cannot find Tcl file: $filename.tcl"
    }
  }

# The following Tcl files must reside in the same directory as
# the present script.

Include constants    ;# Global variables (colors, sizes, etc.)
Include regression   ;# Panel for reporting regression test results.
Include profile      ;# Panel for displaying recognition results.
Include featurestats ;# Interface for examining feature space.
Include thumbnails   ;# Interface for examining training samples.

# Display a message while a command is being performed, then destroy
# the message.  This is used during bootup to display a message while
# training data is being read in.
#
proc ShowProgress { widget message command } {
  global colors font
  label $widget.progress            \
    -background $colors(background) \
    -text $message                  \
    -font $font(normal)             \
    -relief raised                  \
    -borderwidth 3                  \
    -padx 10                        \
    -pady 10
  pack $widget.progress
  update
  eval $command
  destroy $widget.progress
  }

# Create a "toggle" switch between "Recognition" and "Training" modes.  
#
proc MakeToggle {} {
  global toggle font
  frame  .top.toggle       -relief groove -bd 3 -visual best
  label  .top.toggle.mode  -text Mode      -relief ridge -bd 3 -padx 10 -pady 6 -font $font(big)
  button .top.toggle.rec   -text Recognize -command RecognizeMode -width 10 -relief raised -bd 3
  button .top.toggle.group -text Group     -command GroupMode     -width 10 -relief raised -bd 3
  button .top.toggle.train -text Train     -command TrainMode     -width 10 -relief raised -bd 3
  pack   .top.toggle.mode  -side top  -padx 3m -pady 3m
  pack   .top.toggle.rec   -side left -padx 2m -pady 2m
  pack   .top.toggle.group -side left -padx 2m -pady 2m
  pack   .top.toggle.train -side left -padx 2m -pady 2m
  return .top.toggle
  }

# Indicate that the interface is in "Recognition Mode" by changing the
# appearance of the top of the interface, and supplying recogniton-specific
# widgets at the bottom of the interface.
#
proc RecognizeMode {} {
  global toggle train colors
  if { $toggle(mode) != "recognize" } then {
    if { $train(new) > 0 } then QueryPostTrain
    ClearCanvas
   .top.canv         config -highlightbackground $colors(rec)
   .top.toggle.mode  config -background $colors(rec)
   .top.toggle.rec   config -state disabled -relief sunken
   .top.toggle.group config -state normal   -relief raised
   .top.toggle.train config -state normal   -relief raised
   .top.toggle.mode  config -text "Recognition Mode"
    if { $toggle(mode) == "train" } then {destroy .top.train}
    if { $toggle(mode) == "group" } then {destroy .top.group}
    pack [MakeRecInterface] -fill x -expand true -padx 2m -pady 2m 
    set toggle(mode) recognize
    SendMessage set_mode rec
    EnableYesNo 0
    }
  }

# Indicate that the interface is in "Group Mode" by changing the
# appearance of the top of the interface, and supplying grouping-specific
# widgets at the bottom of the interface.
#
proc GroupMode {} {
  global toggle train colors
  if { $toggle(mode) != "group" } then {
    if { $train(new) > 0 } then QueryPostTrain
    ClearCanvas
   .top.canv         config -highlightbackground $colors(group)
   .top.toggle.mode  config -background $colors(group)
   .top.toggle.rec   config -state normal   -relief raised
   .top.toggle.group config -state disabled -relief sunken
   .top.toggle.train config -state normal   -relief raised
   .top.toggle.mode  config -text "Grouping Mode"
    if { $toggle(mode) == "recognize" } then {destroy .top.rec}
    if { $toggle(mode) == "train"     } then {destroy .top.train}
    pack [MakeGroupInterface] -fill x -expand true -padx 2m -pady 2m 
    set toggle(mode) group
    SendMessage set_mode group
    }
  }

# Indicate that the interface is in "Training Mode" by changing the
# appearance of the top of the interface, and supplying training-specific
# widgets at the bottom of the interface.
#
proc TrainMode {} {
  global toggle colors
  if { $toggle(mode) != "train" } then {
    ClearCanvas
   .top.canv         config -highlightbackground $colors(train)
   .top.toggle.mode  config -background $colors(train)
   .top.toggle.rec   config -state normal   -relief raised
   .top.toggle.group config -state normal   -relief raised
   .top.toggle.train config -state disabled -relief sunken
   .top.toggle.mode  config -text "Training Mode"
    if { $toggle(mode) == "group"     } then {destroy .top.group}
    if { $toggle(mode) == "recognize" } then {destroy .top.rec}
    pack [MakeTrainInterface] -fill x -expand true -padx 2m -pady 2m
    SetSampleLabel 0 1
    set toggle(mode) train
    SendMessage set_mode train
    }
  }

# Ask the user if he/she wishes to run "PostTrain" before going back into
# recognition mode (from training mode).  This will make the newly trained
# symbols available to the recognizer.
#
proc QueryPostTrain {} {
  global train
  toplevel .ask_pt -visual best
  set line1 "There are $train(new) new symbols.  These will not be"
  set line2 "recognized until PostTrain is run again."
  set line3 "\nRun PostTrain now?"
  label .ask_pt.msg -text "$line1\n$line2\n$line3"
  frame .ask_pt.row -relief flat -bd 0
  button .ask_pt.row.okay -text Okay -command {
    SendMessage post_proc reload
    set train(new) 0
    destroy .ask_pt
    }
  button .ask_pt.row.can -text Cancel -command {destroy .ask_pt}
  pack .ask_pt.msg      -side top  -padx 8 -pady 8
  pack .ask_pt.row      -side top  -fill x -padx 4 -pady 4
  pack .ask_pt.row.okay -side left -fill x -expand true -padx 4 -pady 4
  pack .ask_pt.row.can  -side left -fill x -expand true -padx 4 -pady 4
  tkwait window .ask_pt
  }

proc MakeRecInterface {} {
  global canv font
  frame  .top.rec         -relief flat -bd 0
  frame  .top.rec.row     -relief flat -bd 0
  label  .top.rec.row.ans -background white -relief raised -bd 3 -text "" -font $font(normal)
  button .top.rec.row.yes -text Yes   -command {SendMessage yes_button; ClearCanvas}
  button .top.rec.row.no  -text No    -command {SendMessage no_button ; ClearCanvas}
  button .top.rec.morph   -text Morph -command {SendMessage morph}

  pack   .top.rec         -padx 4m -pady 4m
  pack   .top.rec.row     -side top -fill x -expand true
  pack   .top.rec.row.no  -side left -padx 1m -pady 1m 
  pack   .top.rec.row.ans -side left -padx 1m -pady 1m -fill both -expand true 
  pack   .top.rec.row.yes -side left -padx 1m -pady 1m 
  pack   .top.rec.morph   -side top -fill x -expand true -padx 1m -pady 2m 
  return .top.rec
  }

proc MakeGroupInterface {} {
  global canv font
  frame  .top.group       -relief flat -bd 0
  button .top.group.clear -text Clear -command {SendMessage group clear}
  button .top.group.done  -text Done  -command {SendMessage group done}

  pack   .top.group       -padx 4m -pady 4m
  pack   .top.group.clear -side left -fill x -expand true -padx 1m -pady 2m 
  pack   .top.group.done  -side left -fill x -expand true -padx 1m -pady 2m 
  return .top.group
  }

proc MakeTrainInterface {} {
  frame  .top.train -relief flat -bd 0
  frame  .top.train.row1  -relief flat -bd 0
  button .top.train.save  -command NameTheSymbol
  frame  .top.train.row2  -relief flat -bd 0

  button .top.train.row1.erase -text Erase  -command {SendMessage erase_sample}
  button .top.train.row1.new   -text "New Sample" -command {SendMessage new_sample}
  button .top.train.row2.left  -text "<<"   -command {SendMessage prev_symbol}
  button .top.train.row2.std   -text "Mark as Exemplar" -command {SendMessage make_exemplar}
  button .top.train.row2.right -text ">>"   -command {SendMessage next_symbol}

  pack .top.train            
  pack .top.train.row1       -side top  -fill x -expand true
  pack .top.train.save       -side top  -fill x -expand true -padx 3 -pady 2
  pack .top.train.row2       -side top  -fill x -expand true
  pack .top.train.row1.erase -side left -fill x -expand true -padx 3 -pady 2
  pack .top.train.row1.new   -side left -fill x -expand true -padx 3 -pady 2
  pack .top.train.row2.left  -side left -fill x -expand true -padx 3 -pady 2
  pack .top.train.row2.std   -side left -fill x -expand true -padx 3 -pady 2
  pack .top.train.row2.right -side left -fill x -expand true -padx 3 -pady 2

  EnableNewErase 0
  SetNumSamples 0
  return .top.train
  }

proc MakeMenuBar {} {
  global font software_version
  # Define the menus.
  frame .top.mbar -relief raised -bd 2
  menubutton .top.mbar.show  -text Show    -underline 1 \
    -menu .top.mbar.show.menu -font $font(normal)
  menubutton .top.mbar.file  -text File    -underline 0 \
    -menu .top.mbar.file.menu -font $font(normal)
  menubutton .top.mbar.spec  -text Special -underline 0 \
    -menu .top.mbar.spec.menu -font $font(normal)

  # Create the pull-down menus associated with the menu buttons.
  menu .top.mbar.show.menu
  menu .top.mbar.file.menu
  menu .top.mbar.spec.menu
  label .top.mbar.version -text "Version $software_version" -relief flat -bd 0

  # Define the entries in the pull-down menus.
  #
  # ---- Show menu ----
  #
 .top.mbar.show.menu add command -label "thumbnails" -command {SendMessage show_thumb ""}
 .top.mbar.show.menu add command -label "profile"    -command {RaiseProfileWindow}
 .top.mbar.show.menu add command -label "features"   -command {SendMessage show_stats on}
 .top.mbar.show.menu add command -label "grouping"   -state disabled
  #
  # ---- File menu ----
  #
 .top.mbar.file.menu add command -label "open symbol"     -state disabled
 .top.mbar.file.menu add command -label "save char"       -state disabled
 .top.mbar.file.menu add command -label "read char"       -state disabled
 .top.mbar.file.menu add command -label "set directory"   -state disabled
  #
  # ---- Special menu ----
  #
  set test_lab "run regression test"
  set dump_lab "dump current sample"
  set exit_lab "exit (no post-processing)"
 .top.mbar.spec.menu add command -label "$test_lab" -command {SendMessage run_reg_test ""}
 .top.mbar.spec.menu add command -label "$dump_lab" -command {SendMessage print_lines}
 .top.mbar.spec.menu add command -label "$exit_lab" -command {SendMessage quit force}
 .top.mbar.spec.menu add command -label "quit"      -command {SendMessage quit}

  # Pack the objects in the menu bar.
  pack .top.mbar.show .top.mbar.file .top.mbar.spec -side left
  pack .top.mbar.version -side right -padx 4
  return .top.mbar
  }

proc PostAnswer { ans } { 
  global toggle
  if { $toggle(mode) != "recognize" } return
 .top.rec.row.ans config -text "$ans" 
  }

proc ClearCanvas { {flag ""} } {
  global toggle
 if { $flag == "" } then {SendMessage clear}
 .top.canv delete ALL
 .top.canv delete LABEL
  if { $flag == "" && $toggle(mode) == "recognize" } then { 
    PostAnswer ""
    EnableYesNo 0    
    }
  }

proc SavePostscript { {file_name ""} } {
  puts "\n Writing $file_name ..."
  set fileid [open $file_name w]
  puts $fileid [.top.canv postscript]
  close $fileid
  }

# Draw Lines in the canvas
proc DrawLine { x1 y1 x2 y2 {color black} } {
 .top.canv create line $x1 $y1 $x2 $y2 -width 3 -tag ALL -fill $color
  }

proc DrawGroupBox { xmin ymin xmax ymax {color black} } {
 .top.canv create rectangle $xmin $ymin $xmax $ymax -outline $color -tag ALL
  }

# Draw a filled polygon.
proc FillPolygon { coords {color black} } {
  eval ".top.canv create polygon $coords -fill $color -tags {ALL PGON}"
 .top.canv lower PGON ALL
  }

# Draw the line from the last point to the current point
proc move_to { x y } {
  global coord timer
  set coord(x) $x
  set coord(y) $y
  SendMessage move_to $x $y
  if { $timer(id) != "" } then { after cancel $timer(id) }
  }

# Draw the line from the last point to the current point
proc draw_to { x y } {
  global coord canv
 .top.canv create line $coord(x) $coord(y) $x $y -width 3 -tag ALL
  set coord(x) $x
  set coord(y) $y
  SendMessage draw_to $x $y
  }

proc end_line {} {
  global timer
  SendMessage end_line
  set timer(id) [after $timer(delay) TimeOut]
  }

# This is invoked when the user has stopped drawing for a specified
# period of time.
proc TimeOut {} {
  global timer
  set timer(id) ""
  SendMessage timed_out
  }

# Display an error messge in a pop-up window with an "okay" button.
proc Error_Message { message } {
  global font
  if [catch "toplevel .error -visual best"] then return
  message  .error.msg -justify center -relief sunken -bd 2 -text $message -width 5c -font $font(normal)
  button   .error.ok  -text "OK" -command {destroy .error}
  pack     .error.msg .error.ok -side top -fill x -expand true -padx 10 -pady 5
  }

proc NameTheSymbol {} {
  global SymbName colors font
  set SymbName ""
  toplevel .save
  label    .save.label  -text "Enter symbol name"
  entry    .save.entry  -width 25 -relief sunken -bd 2 -textvariable SymbName \
                        -background $colors(entry) -font $font(normal)
  frame    .save.row    -relief flat -bd 0
  button   .save.row.cancel -text Cancel -relief raised -command { destroy .save }
  button   .save.row.save   -text Save   -relief raised -command {Save_it}
  pack     .save.row.cancel -side left -padx 2 -pady 2 -fill x -expand true
  pack     .save.row.save   -side left -padx 2 -pady 2 -fill x -expand true
  pack     .save.label .save.entry .save.row -side top -padx 2m -pady 1m -fill x -expand true
  bind     .save.entry <KeyPress-Return> Save_it
  }

proc Save_it {} {
  global SymbName train
  if { $SymbName == "" } then { Error_Message "Symbol name must be specified" } else {
    incr train(new) 
    destroy .save
    SendMessage save_samples $SymbName
    ClearCanvas
    }
  }

proc SetNumSamples { n } {
  if { $n == 0 } then { set state disabled; set lab "Save samples"    }
  if { $n == 1 } then { set state normal  ; set lab "Save 1 sample"   }
  if { $n  > 1 } then { set state normal  ; set lab "Save $n samples" }
 .top.train.save config -text $lab -state $state
  update
  }

proc SetNumSymbols { n } {
  global numsym
  set numsym(total) $n
 .top.numsym config -text "$numsym(label) $numsym(total)"
  update
  }

proc SetSampleLabel { curr {exemplar 0} } {
  global font colors canv
 .top.canv delete LABEL
 .top.canv create text 10 5 -text "sample #[expr $curr + 1]" -font $font(bold) \
    -fill $colors(lite-gray) -tag LABEL -anchor nw -justify left
  if { $exemplar != "0" } then {
   .top.canv create text $canv(main-width) 5 -text "exemplar" -font $font(bold) \
      -fill $colors(lite-red) -tag LABEL -justify right -anchor ne
    }
  }

proc EnableMorph { {on 1} } { 
  global toggle
  if { $toggle(mode) != "recognize" } return
  if { $on == 1 } then {
   .top.rec.morph   config -state normal
  } else {
   .top.rec.morph   config -state disabled
    }
  }

proc EnableYesNo { {on 1} } { 
  global toggle
  if { $toggle(mode) != "recognize" } return
  if { $on == 1 } then {
   .top.rec.row.no  config -state normal
   .top.rec.row.yes config -state normal
   .top.rec.morph   config -state normal
  } else {
   .top.rec.row.no  config -state disabled
   .top.rec.row.yes config -state disabled
   .top.rec.morph   config -state disabled
    }
  }

proc EnableNewErase { {on 1} } {
  if { $on == 1 } then {
   .top.train.row1.new   config -state normal
   .top.train.row1.erase config -state normal
  } else {
   .top.train.row1.new   config -state disabled
   .top.train.row1.erase config -state disabled
   .top.train.row2.std   config -state disabled
    }
  }

proc EnableExemplar { {on 1} } {
  if { $on == 1 } then {
   .top.train.row2.std config -state normal
  } else {
   .top.train.row2.std config -state disabled
    }
  }

proc ConfirmQuit {} {
  global font
  toplevel .confirm
  label    .confirm.label  -text "There are unsaved samples.\nReally Quit?"
  frame    .confirm.row    -relief flat -bd 0
  button   .confirm.row.confirm -text "Really Quit" -relief raised -command {SendMessage quit force}
  button   .confirm.row.cancel  -text "Cancel"       -relief raised -command {destroy .confirm }
  pack     .confirm.row.confirm -side left -padx 2 -pady 2 -fill x -expand true
  pack     .confirm.row.cancel  -side left -padx 2 -pady 2 -fill x -expand true
  pack     .confirm.label .confirm.row -side top -padx 2m -pady 1m -fill x -expand true
  }


proc old-SendMessage { msg {arg1 ""} {arg2 ""} {arg3 ""} } {
  puts "$msg $arg1 $arg2 $arg3"
  }

proc MakeMainInterface {} {
  global font canv numsym colors
  pack [MakeMenuBar] -side top -fill x 
  label .top.numsym -text "$numsym(label) $numsym(total)" -relief flat -font $font(normal)
  pack  .top.numsym -side top -padx 1m -pady 2m
  pack [MakeToggle] -padx 1m -pady 1m
  canvas .top.canv             \
    -width $canv(main-width)   \
    -height $canv(main-height) \
    -background white          \
    -relief flat               \
    -borderwidth 0             \
    -highlightthickness 3      \
    -highlightbackground $colors(rec)
  pack .top.canv  -side top  -padx 5m -pady 1m
  bind .top.canv <ButtonPress>   {move_to %x %y}
  bind .top.canv <B1-Motion>     {draw_to %x %y}
  bind .top.canv <ButtonRelease> {end_line}
  RecognizeMode
  }

proc InteractiveMode {} {
  global colors
  frame .top -visual best
  pack .top
  ShowProgress .top "Loading feature matrices..." {
    SendMessage load_data
    MakeMainInterface
    }
  }

proc RegressionOnly {} {
  global colors
  frame .top -visual best
  pack .top
  ShowProgress .top "Loading feature matrices..." {
    SendMessage load_data
    }
  SendMessage run_reg_test .top
  }

proc ThumbnailsOnly {} {
  global colors
  frame .top -visual best
  pack .top
  ShowProgress .top "Loading feature matrices..." {
    SendMessage load_data
    }
  SendMessage show_thumb .top
  }

if { $startup_mode == "interactive_mode" } then { InteractiveMode }
if { $startup_mode == "regression_only"  } then { RegressionOnly  }
if { $startup_mode == "thumbnails_only"  } then { ThumbnailsOnly  }

tkwait window .








