/***************************************************************************
* Features.h                                                               *
*                                                                          *
* Defines classes for creating and managing vectors of features and their  *
* statistics.  These classes are very fundamental to feature-based         *
* recognition.                                                             *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/15/99  arvo    Removed fixed-length arrays.                     *
*       12/21/97  arvo    major cleanup.                                   *
*       01/05/97  guo     Initial codiing.                                 *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __FEATURES_INCLUDED__
#define __FEATURES_INCLUDED__

#include <assert.h>
#include <iostream.h>
#include <math.h>
#include <stdio.h>
#include <util/Math.h>
#include <util/Multiline.h>
#include <util/List.h>       // Used for chaining letter statistics together.
#include <rec/Symbol.h>      // The Symbol class links names to typefaces.
#include <rec/Config.h>

class FeatureVector;      // The set of all features for one symbol.
class FeatureStats;       // Statistics for one feature of one symbol.
class FeatureStatsVector; // Statistics for all features of one symbol.


// The FeatureVector holds all the features (which are single floating
// point numbers) computed for a specific hand-written symbol.  It is 
// essentially a point in feature space.  Features can be conveniently 
// added using the "<<" operator, so that keeping track of indices is 
// unnecessary.  This vector has a fixed length because the number of 
// features is fixed at compile time.  The "<<" operator can also be
// used to insert comments, so that the meanings of the features can
// be looked up later.
 
class FeatureVector : public ListItem {
    public:
        FeatureVector() { Clear(); }
       ~FeatureVector() {}
        inline FeatureVector &operator<<( float );
        inline FeatureVector &operator<<( const Vec2 & );
        inline FeatureVector &operator<<( const char * );
        float  Feature( int i ) const { return feature[i]; }
        const char *Comment( int i ) const;
        int   Size () const { return index; }
        void  Clear();
    private:
        int index;
        float feature[ Max_Features ];
        const char *comment[ Max_Features ];
    };


// The FeatureStats structure holds all of the statistics used to
// characterize a single feature of a collection of hand-written
// symbol (written in a single style).  

class FeatureStats {
    public:
        FeatureStats();
       ~FeatureStats() {};
        void Clear();
        float Average;
        float StndDev;
        float Min;
        float Max;
    };


// The FeatureStatsVector holds the result of analyzing a collection of
// hand-written symbols, all representing the same symbol written in
// the same style.  This vector has a fixed length because the number of
// features is fixed at compile time.  Instances of this class can be 
// chained together to form a list; this feature is used to create a table 
// of all recognized symbols.
 
class FeatureStatsVector : public ListItem {
    public:
        FeatureStatsVector();
       ~FeatureStatsVector() {};
        FeatureStatsVector &operator<<( const FeatureStats & );
        const FeatureStats &Stats( int i ) const { return stats[i]; }
        Symbol *GetSymbol  () const { return symbol;   }
        float   GetDistance() const { return distance; }
        void    SetSymbol  ( Symbol *symb ) { symbol   = symb; }
        void    SetDistance( float   dist ) { distance = dist; }
        void    ProcessTrainingFile( const char *file_name );
        int  Size () const { return index; }
        void Clear();
        void Close(); // Finish processing the statistics.
        void operator+=( const FeatureVector & );
        static const FeatureStatsVector Null;
    private:
        int     index;
        int     num_samples;
        Symbol *symbol;
        float   distance;
        FeatureStats stats[ Max_Features ];
    };


// I/O operators for the classes defined in this header.

extern istream &operator>>( istream &,       FeatureStats       & );
extern ostream &operator<<( ostream &, const FeatureStats       & );
extern istream &operator>>( istream &,       FeatureStatsVector & );
extern ostream &operator<<( ostream &, const FeatureStatsVector & );


// Inline methods for adding values and comments to a feature vector.

inline FeatureVector &FeatureVector::operator<<( float x )
    {
    assert( index < Max_Features );
    feature[ index++ ] = x;
    return *this;
    }

inline FeatureVector &FeatureVector::operator<<( const Vec2 &v )
    {
    assert( index < Max_Features - 1 );
    feature[ index++ ] = v.X();
    feature[ index++ ] = v.Y();
    return *this;
    }

inline FeatureVector &FeatureVector::operator<<( const char *s )
    {
    comment[ index ] = s;
    return *this;
    }


// ExtractFeatures computes the feature vector of a specific collection
// of hand-written strokes the represent a single character.  The strokes
// are encoded as a single Multiline.  This vector is used to find the K 
// nearest neighbors in feature space.  Computing this vector is the first 
// step in the recognition process.  It is also used to compute the 
// FeatureStatsVector, which is stored in a feature matrix file.

extern void ExtractFeatures( 
    const Multiline &lines, 
    FeatureVector &FV
    );

#endif



