/***************************************************************************
* Recognizer.h                                                             *
*                                                                          *
* Basic class definitions for symbol recognizers and the "profiles" that   *
* are generated by recognizers.                                            *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       10/22/99  arvo    Removed modifier "Generic" from class names.     *
*       08/31/99  arvo    Added Uncertainty member to Profile.             *
*       08/15/99  arvo    Removed fixed-length arrays.                     *
*       11/13/97  arvo    Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __RECOGNIZER_INCLUDED__
#define __RECOGNIZER_INCLUDED__

#include <iostream.h>
#include <util/Multiline.h>
#include <rec/Symbol.h>

class Recognizer; // Base class for all symbol recognizers.
class Profile;    // Base class for all recognition profiles.


// The maximum number of choices that a recognizer can return as
// part of a "profile" is limited to the following number.  Any 
// recognizer may choose to return fewer for any given symbol.

static const int MaxProfileChoices = 10;


// The RecStatusType lists a number of generic error codes that may 
// be returned by a recognizer.  Other codes may also be returned by a
// recognizer, but should be offset by "rec_other_error" to avoid
// confusion with the generic ones.

enum RecStatusType {
    rec_undefined,
    rec_okay,
    rec_unrecognized,
    rec_file_open_error,
    rec_file_stat_error,
    rec_version_error,
    rec_internal_error,
    rec_other_error     // Add this offset to additional error codes.
    };


// Each recognizer returns a "profile," which encodes a probability 
// distribution of possible matches for a given hand-drawn character, 
// organized from most to least likely.  A recognizer may use this 
// generic (base) class directly, or define a custom profile class by 
// inheriting from this one and overriding one or more of the virtual 
// methods.

class Profile {

    public:
      Profile() { Clear(); }
      Profile( const Profile &prof ) { *this = prof; }
      virtual ~Profile() {}
      Profile &operator=( const Profile & );
      Symbol *BestSymbol () const { return GetSymbol(0); }
      float   BestProb   () const { return GetProb  (0); }
      int     NumChoices () const { return num_choices ; }
      float   Uncertainty( const char *symbname = NULL ) const;
      int     GetOrdinal( const char   *name ) const;
      int     GetOrdinal( const Symbol *lett ) const;

    public: // Virtual methods that may be redefied by derived profile classes.
      virtual  Symbol *GetSymbol ( int i ) const { return S[i]; }
      virtual  float   GetProb   ( int i ) const { return P[i]; }
      virtual  void    SetSymbol ( int i, Symbol *sym  ) { S[i] = sym; }
      virtual  void    SetProb   ( int i, float   prob ) { P[i] = prob; }
      virtual  void    SetChoices( int num );
      virtual  void    Normalize ();

    protected:
      int     num_choices;
      Symbol *S[ MaxProfileChoices ];
      float   P[ MaxProfileChoices ];
      void    Clear();

    };


// The "Recognizer" class is a purely virtural class that serves only
// as the base for symbol recognizers.  Each recognizer must define a new 
// class by deriving from this one and overriding at least the "Recognize"
// virtual method, and possibly others (and, of course, by adding any 
// appropriate private data members).  A recognizer typically reads training 
// data from a file or a directory whose path name is (optionally) provided
// in the constructor.  Then, the "Recognize" function returns a probability 
// distribution, or "profile", for user-drawn symbols that are encoded as 
// multilines (i.e. a collection of one ore more strokes).

class Recognizer {

    public:
      Recognizer( const char * = NULL ) {}
      virtual ~Recognizer() {}

    public: // This purely virtual method MUST be supplied.
      virtual int Recognize( const Multiline &, Profile & ) = 0;

    public: // These virtual methods are optional.
      virtual int LoadTrainingFiles( const char * ) { return rec_okay; }
      virtual int Clear() { return rec_okay; }
      virtual int NumSymbols() const { return 0; }

    };


// I/O operators for the classes defined in this header.

extern ostream &operator<<( ostream &out, const Profile & );
extern istream &operator>>( istream &in ,       Profile & );

#endif
