/***************************************************************************
* Box2.h                                                                   *
*                                                                          *
* Basic 2D bounding boxes.                                                 *
*                                                                          *
*   HISTORY                                                                *
*       05/12/98    arvo    Broken out into a separate module.             *
*       12/04/97    arvo    Reorganized for use as a library.              *
*       07/09/96    arvo    Initial coding.                                *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __BOX2_INCLUDED__
#define __BOX2_INCLUDED__

#include <stream.h>
#include <util/Vec2.h> // 2-D vectors.

// 2-Dimensional Bounding box with associated geometric operations.

class Box2d {

    public:
        Box2d();
        Box2d( const Vec2 &A, const Vec2 &B ) { Set( A, B ); }
        Box2d( const Box2d &B ) { *this = B; }
       ~Box2d() {}
        Box2d &operator+=( const Vec2  & );  // Translate the box.
        Box2d &operator-=( const Vec2  & );  // Translate the box.
        Box2d &operator*=( float         );  // Scale the box.
        Box2d &operator^=( float         );  // Centered scale.
        Box2d &operator<<( const Vec2  & );  // Expand to include this point.
        Box2d &operator<<( const Box2d & );  // Expand to include this box.
        Box2d &operator*=( const Box2d & );  // Intersection of boxes.

    public:
        Box2d  operator* ( const Box2d & ) const; // Intersection of boxes.
        Box2d  operator* ( float         ) const; // Scaled box.
        Box2d  operator+ ( const Vec2  & ) const; // Translated box.
        Box2d  operator- ( const Vec2  & ) const; // Translated box.
        Box2d  Expand    ( float delta   ) const; // expand max, shrink min.
        float  dist      ( const Vec2  & ) const; // Euclidean dist from point to box.
        float  dist      ( const Box2d & ) const; // Euclidean dist from box to box.
        Vec2   Closest   ( const Vec2  & ) const; // Point on box closest to point.
        Vec2   Closest   ( const Box2d & ) const; // Vector from box to closest point.
        int    Inside    ( const Vec2  & ) const; // 1 if point is in box.
        float  Xmin      () const { return xmin; }
        float  Xmax      () const { return xmax; }
        float  Ymin      () const { return ymin; }
        float  Ymax      () const { return ymax; }
        float  Width     () const { return xmax - xmin; }
        float  Height    () const { return ymax - ymin; }
        float  Diam      () const;
        float  Area      () const;
        float  Perim     () const;
        Vec2   Min       () const { return Vec2( xmin, ymin ); }
        Vec2   Max       () const { return Vec2( xmax, ymax ); }
        Vec2   NE        () const { return Vec2( xmax, ymax ); }
        Vec2   NW        () const { return Vec2( xmin, ymax ); }
        Vec2   SE        () const { return Vec2( xmax, ymin ); }
        Vec2   SW        () const { return Vec2( xmin, ymin ); }
        Vec2   Diag      () const { return Vec2( Width(), Height() ); }
        Vec2   Center    () const { return 0.5 * Vec2( xmin + xmax, ymin + ymax ); }

        const char *ToString( char *buff = NULL ) const;
        Box2d &FromString( const char *str );
        Box2d &Set       ( const Vec2 &, const Vec2 & );

    public:
        static const Box2d Huge; // An "infinite" box.
        static const Box2d Null; // The default: maximally "degenerate".
        static const Box2d Zero; // All elements are zero.

    private:
        Box2d( float size ); 
        float xmin, xmax;
        float ymin, ymax;
    };


extern float dist( 
    const Box2d &, 
    const Box2d & 
    );

extern int hit( 
    const Box2d &, 
    const Box2d & 
    );

extern ostream &operator<<( 
    ostream &out, 
    const Box2d & 
    );

#endif
