/***************************************************************************
* FeatureRecognizer.C                                                      *
*                                                                          *
* Class implementation for feature-based symbol recognizer.                *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/06/99  arvo    Significant restructuring & bug fixing.          *
*       04/06/98  fedorov Changed to work without SymbolNames              *
*       02/10/98  fedorov Recoded to use multiline                         *
*       11/20/97  xfan    Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <fstream.h>
#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include <stream.h>
#include <string.h>
#include <util/File.h>
#include <util/Char.h>
#include <rec/Font.h>
#include <rec/Features.h>
#include <rec/FeatureRecognizer.h>

const FeatureStatsVector FeatureStatsVector::Null;
static int debug = 0;

/***************************************************************************
* FeatureRecognizer
*
* The FeatureRecognizer class is a sub-class of Recognizer.  It implements
* the abstract recognizer class using a nearest-neighbor algorithm in 
* feature space.
* 
****************************************************************************/
FeatureRecognizer::FeatureRecognizer( const char *path )
    {
    num_symbols  = 0;
    num_features = 0;
    prog_meter   = NULL;
    if( path != NULL ) LoadFeatureMatrices( path );
    }

FeatureRecognizer::~FeatureRecognizer()
    {
    }

/***************************************************************************
* LoadFeatureMatrices
*
* A "FeatureMatrix" is a file that contains statistics generated from one or
* more collections of multilines, each collection representing samples of a
* single hand-drawn character.  It is the feature matrices that are used by
* the recognizer in determining distances in feature space.  The format of
* a feature matrix file is as follows:
*
*    [Feature Matrix: Version <version-number>]
*    <number-of-symbols>
*
*    <symbol-name> <training-file-name>
*    1: <mean> <std-dev> <min> <max>
*    2: <mean> <std-dev> <min> <max>
*    3: <mean> <std-dev> <min> <max>
*    4: <mean> <std-dev> <min> <max>
*    ....
*
*    <symbol-name> <training-file-name>
*    1: <mean> <std-dev> <min> <max>
*    2: <mean> <std-dev> <min> <max>
*    3: <mean> <std-dev> <min> <max>
*    4: <mean> <std-dev> <min> <max>
*    ....
*                       
* Feature matrices are generated from training files by PostTrain.
*                       
****************************************************************************/
int FeatureRecognizer::LoadFeatureMatrices( const char *dir_name )
    {
    int status;
    Directory tree( dir_name, D_Files | D_Dirs );

    // Initialize everything so that it can be rebuilt from scratch, in
    // case this function is called mid-way through an interactive session.
    // FontTable::FT.Clear();
    feature_stats_list.Clear();
    num_symbols  = 0;
    num_features = 0;

    // Loop through all elements of the directory looking for subdirectories
    // and feature matrices.  Recursively explore the subdirectories, and
    // load the feature matrices.  Ignore all other files.  The order in
    // which the feature matrices are found is immaterial.
    while( tree.Next() )
        {
        if( tree == "FeatureMatrix" )
            {
            status = LoadFeatureMatrix( tree.FullPath() );
            if( status != rec_okay ) return status;
            }
        }
    return rec_okay;
    }

/***************************************************************************
* CheckFeatureMatrixHeader
*
* Check that the first two lines of a FeatureMatrix file look like this:
*
*    [Feature Matrix: Version <version-number>]
*    <number-of-symbols>
*
* If so, check that the version number matches the version number of the
* code.  If everything checks out, return the number of symbols as the 
* function value, otherwise return -1.
* 
****************************************************************************/
static int CheckFeatureMatrixHeader( istream &fmat, const char *path )
    {
    char  in_buffer[256];
    int   nsymbols;
    float version = 0;

    fmat.getline( in_buffer, sizeof(in_buffer) );
    int count = sscanf( in_buffer, FM_VersionFormat, &version );

    if( count != 1 )
        {
        cerr << "\nERROR: file " << path << " does not have the correct format." << endl;
        return -1;
        }

    if( version != FM_Version )
        {
        cerr << "\nERROR: file "  << path       << " is out of date. "
             << "Found version "  << version    << " "
             << "when expecting " << FM_Version << "." << endl
             << "Run PostTrain again to update FeatureMatrix." << endl;
        return -1;
        }

    fmat >> nsymbols;
    return nsymbols;
    }

/***************************************************************************
* LoadFeatureMatrix
*
* Load statistics from the "FeatureMatrix" file into the "feature_stats_list"
* list.  This is the main list contining the statistics for all symbols that
* the recognizer will recognize.  Each feature matrix may contain the
* statistics for an arbitrary number of symbols.
* 
****************************************************************************/
int FeatureRecognizer::LoadFeatureMatrix( const char *path )
    {
    char message[256];

    filebuf fbuff;
    if( fbuff.open( path, ios::in ) == NULL )
        {
        cerr << "Could not open training file " << path << endl;
        return rec_file_open_error;
        }
    istream fmat( &fbuff );

    int symbols_in_file = CheckFeatureMatrixHeader( fmat, path );
    if( symbols_in_file < 0 ) return rec_version_error;

    // Initialize the list "feature_stats_list":
    // Create a new "FeatureStatsVector" instance for each symbol encoded in 
    // the feature matrix file, fill it with the statistics following the symbol
    // name in the file, and store all such FeatureStatsVector instances in a 
    // list (feature_stats_list) that will be consulted when recognizing each 
    // new character.
    //
    for( int i = 0; i < symbols_in_file; i++ ) 
        {
        // Create and initialize a new FeatureStatsVector instance, and insert 
        // it into the growing list of symbols that are to be recognized.
        FeatureStatsVector *fsv = new FeatureStatsVector;
        feature_stats_list << fsv;

        // Read the FeatureStatsVector using its input operator.  Check that
        // all vectors have the same length, and keep this number for future
        // queries.
        fmat >> *fsv;
        if( num_features == 0 ) num_features = fsv->Size();
        else assert( fsv->Size() == num_features );

        // Generate a message for the progress meter.  This is not used by
        // the recognizer.
        Symbol *symbol = fsv->GetSymbol();
        sprintf( message, "%3d [%s] %s", num_symbols, symbol->Name(), symbol->Qual() );
        Progress( 1, message );

        num_symbols++;
        }
    fbuff.close();
    return rec_okay;
    }

/***************************************************************************
* GetExemplar
*
* Open the given training file and read the up to and including the first
* Multiline.  Return the name of the symbol and the first Mutiltiline, which
* is the exemplar of the training data.
* 
****************************************************************************/
int FeatureRecognizer::GetExemplar( const char *fname, Multiline &exemp, char *name ) const
    {
    char in_buffer[256];
    char temp_name[64];
    int  num_samples;

    filebuf fbuff;
    fbuff.open( fname, ios::in );
    istream in( &fbuff );

    if( !in ) return 0;
    // Skip over the header information.
    in.getline( in_buffer, sizeof(in_buffer) );

    in >> num_samples;
    in >> temp_name;
    if( num_samples > 0 ) in >> exemp;

    fbuff.close();
    if( name != NULL ) strcpy( name, temp_name );
    return 1;
    }

/***************************************************************************
* GetBestCandidates
*
* This function returns the indices of the K smallest distances in the 
* FeatureStatsVector, which were computed by ComputeFeatureDistances.
* 
****************************************************************************/
int FeatureRecognizer::GetBestCandidates( int top_n, FeatureStatsVector *Index[] )
    {
    ListItem *item;
    for( int i = 0; i < top_n; i++ ) Index[i] = NULL;
    for( item = feature_stats_list.First(); item != NULL; item = item->Next() )
        {
        FeatureStatsVector *fsv = (FeatureStatsVector *)item;
        // Compare this distance against all others saved so far.
        for( int j = 0; j < top_n; j++ )
            {
            FeatureStatsVector *ind = Index[j];
            // Add the new index if the distance is smaller than one of the
            // distances currently in the list, or if we don't have all n
            // of them yet.
            if( ind == NULL || fsv->GetDistance() < ind->GetDistance() )
                {
                // Move the others down to make room for the new one.
                for( int k = top_n - 1; k > j; k-- )
                    {
                    Index[k] = Index[k-1];
                    }
                Index[j] = fsv;
                break;
                }
            }
        }
    // Return the number of candidates actually found.
    for( int j = 0; j < top_n; j++ ) if( Index[j] == NULL ) return j;
    return top_n;
    }

/***************************************************************************
* ComputeFeatureDistances
*
* This function computes the feature vector of the given Multiline and
* finds its distance from all other feature vectors defined by other symbols
* in feature space.  The results are left in the FeatureStatsVector and
* a pointer to the closest point is returned as the function value.
* 
****************************************************************************/
FeatureStatsVector *FeatureRecognizer::ComputeFeatureDistances( const Multiline &lines )
    {
    FeatureVector F;
    ExtractFeatures( lines, F );

    FeatureStatsVector *best = 0;
    double Min = 1.0E10;
    for( ListItem *item = feature_stats_list.First(); item != NULL; item = item->Next() )
        {
        FeatureStatsVector *fsv = (FeatureStatsVector *)item;
        register double Sum = 0;
        for( register int j = 0; j < fsv->Size(); j++ )
            {
            const FeatureStats &stats = fsv->Stats(j);
            float feature = F.Feature(j);
            float width   = stats.Max - stats.Min;
            if( stats.StndDev != 0 ) Sum += Abs( (stats.Average - feature) / stats.StndDev );
            if( feature < stats.Min - width ) Sum += 1.0; else
            if( feature < stats.Min         ) Sum += 0.5; else
            if( feature > stats.Max + width ) Sum += 1.0; else
            if( feature > stats.Max         ) Sum += 0.5;
            }

        // if( F.Feature(0) != fsv->Stats(0).Average ) Sum += 100;

        fsv->SetDistance( Sum );

        if( Sum < Min ) 
            {
            Min  = Sum;
            best = fsv;
            }
        }
    return best;
    }

/***************************************************************************
* Recognize        -*-*- THE MAIN FUNCTION OF THE RECOGNIZER -*-*-
*
* This is the main function of the recognizer: It tales a collection of
* strokes drawn by the user and finds the K closest symbols in feature space.
* These K symbols are returned in the "Profile", sorted by decreasing
* confidence.
* 
****************************************************************************/
int FeatureRecognizer::Recognize( const Multiline &lines, Profile &prof )
    {
    static const int N = 30;
    static FeatureStatsVector *Index[N];

    // Compute distances to all points in feature space.  This computation
    // is O(nk), where n is the number of symbols, and k is the number 
    // of features.
    FeatureStatsVector *best = ComputeFeatureDistances( lines );

    if( debug )
        {
        ListItem *item = feature_stats_list.First(); 
        while( item != NULL )
            {
            FeatureStatsVector *fsv = (FeatureStatsVector *)item;
            cerr << "Symbol "      << fsv->GetSymbol()->Name() << ", "
                 << "Confidence: " << fsv->GetDistance() << endl;
            item = item->Next();
            }
        cerr << "Recognized as Symbol " << best->GetSymbol()->Name() << endl;
        }

    int actual_choices = GetBestCandidates( Min( MaxProfileChoices, N ), Index );
    prof.SetChoices( actual_choices );

    // Copy the top few choices into the profile that was passed in.
    // Here is where distances get mapped into confidences, so that
    // large scores mean higher confidence (or probability).  Currently,
    // there is no guarantee that the score will be less than 1 (although
    // it typically is).
    for( int i = 0; i < actual_choices; i++ ) 
        {
        FeatureStatsVector *fsv = Index[i];
        prof.SetSymbol( i, fsv->GetSymbol() );
        prof.SetProb  ( i, fsv->Size() / ( 1.0 + fsv->GetDistance() ) );
        }

    return rec_okay;
    }

/***************************************************************************
* ProcessTrainingFile
*
* This function fills in the statistics of a FeatureStatsVector instance with
* statistics generated from the samples of a specific training file.  The 
* statistics are computed for each element of the feature vector.  Progress 
* is printed to standard out.
* 
****************************************************************************/
void FeatureStatsVector::ProcessTrainingFile( const char *file_name )
    {
    FeatureVector vec;
    Multiline lines;
    int   num_samples;
    char  symb_name[64];
    char  in_buffer[256];
    float version = 0;

    filebuf fbuff;
    fbuff.open( file_name, ios::in );
    istream in( &fbuff );

    if( !in ) return;
    in.getline( in_buffer, sizeof(in_buffer) );
    int count = sscanf( in_buffer, TD_VersionFormat, &version );
      
    if( count != 1 ) 
        {
        cerr << "ERROR: FIle " << file_name 
             << " is not a valid training file." << endl;
        return;
        }
    else if( version < TD_Version ) 
        {
        cerr << "ERROR: The training file " << file_name 
             << " is an obsolete format." << endl;
        return;
        }
    else if( version > TD_Version ) 
        {
        cerr << "ERROR: The training file " << file_name 
             << " is a newer than this software can handle." << endl;
        return;
        }

    in >> num_samples;
    in >> symb_name;
     
    cout << form( "symbol %s, (%d samples)  ", symb_name, num_samples );

    // Initialize the local fields of the FeatureStatsVector instance.
    distance = 0.0;
    symbol   = FontTable::FT.Insert( symb_name, file_name );

    // Read all the multilines from the file, compute their corresponding
    // feature vectors, and accumulate statistics for each feature of the
    // vector in the FeatrueStatsVector.
    //
    FeatureStatsVector::Clear();
    for( int samp = 0; samp < num_samples; samp++ )
        {
        // Now read all the samples.
        in >> lines;
        if( !in )
            {
            cerr << form( "ERROR reading sample %d ", samp + 1 )
                 << form( "out of %d ", num_samples )
                 << form( "in file %s", file_name )
                 << endl;
            exit(1);
            }
        ExtractFeatures( lines, vec );
        *this += vec;
        } 
    // Finish the computation of the statistics accumulated via the += operator.
    FeatureStatsVector::Close(); 

    cout << endl;
    fbuff.close();
    }

/***************************************************************************
* GetStats (returns FeatureStatsVector of a symbol)
*
* The purpose of this function is to look up the collection of statistics
* associated with a given symbol and return a const reference to it.  This
* is strictly for visualization of the statistics -- it has nothing to do
* with the actual recognition process.
* 
****************************************************************************/
const FeatureStatsVector &FeatureRecognizer::GetStats( Symbol *symbol ) const
    {
    ListItem *item;
    for( item = feature_stats_list.First(); item != NULL; item = item->Next() )
        {
        FeatureStatsVector *fsv = (FeatureStatsVector *)item;
        if( fsv->GetSymbol() == symbol ) return *fsv;
        }
    return FeatureStatsVector::Null; // Symbol not found.
    }


void FeatureRecognizer::Progress( int flag, const char *msg )
    {
    if( prog_meter != NULL ) prog_meter( flag, msg );
    }




