/***************************************************************************
* PostTrain.C                                                              *
*                                                                          *
* Functions for generating "Feature Matrix" files from raw training data.  *
* This must be performed after new training files have been added or       *
* existing ones have been modified, since the feature recognition code     *
* reads only the feature matrices, not the training data.                  *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/15/99     arvo    Removed all fixed-length arrays.              *
*       08/06/99     arvo    Major cleanup.                                *
*       01/05/97     guo     Initial codiing.                              *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <stream.h>
#include <assert.h>
#include <util/Char.h>
#include <util/List.h>
#include <util/File.h>
#include <util/Args.h>
#include <rec/Font.h>
#include <rec/PostTrain.h>

static int 
    total_symbols  = 0,
    total_matrices = 0;

static void tab( int level )
    {
    for( int i = 0; i < level; i++ ) cout << "  ";
    }

/***************************************************************************
* Given a list of FeatureStatVectors, this function writes the information
* to a file called "FeatureMatrix" in the given directory.  The FeatureMatrix
* is intended to encode all the information embodied in the training files
* found within the same directory.  Thus, the feature-based recognizer need
* only read the FeatureMatrix files, and not the raw training data.
***************************************************************************/
void WriteFeatureMatrix( const char *dir_name, const List &feature_stats, int level )
    {
    filebuf fbuff;
    char filename[ Max_Name_Length ];
    int  entries = feature_stats.NumEntries();

    sprintf( filename, "%s/FeatureMatrix", dir_name );
    fbuff.open( filename, ios::out );
    ostream out( &fbuff );

    out << form( FM_VersionFormat, FM_Version ) 
        << entries << endl 
        << endl;

    for( ListItem *item = feature_stats.First(); item != NULL; item = item->Next() ) 
        {
        FeatureStatsVector *fsv = (FeatureStatsVector *)item;
        out << *fsv;
        out << endl;
        }

    tab( level );
    cout << form( "Encoded %d symbols.\n\n", entries );
    total_symbols  += entries;
    total_matrices += 1;
    fbuff.close();
    }

/***************************************************************************
* Read in all training files in each directory and compute a "FeatureMatrix"
* file that captures the characterizes the statistical variation of each
* feature within the samples of each symbol (and style); these feature 
* matrices are in turn used by the recognizer, rather than the original 
* training data.  A FeatureMatrix represents all the training files found
* in a SINGLE directory; thus, many feature matrices may be generated by
* this function -- one for each sub-directory.
***************************************************************************/
static void ComputeFeatureMatrix( const char *dir_name, int level = 1 )
    {
    char buff[256];
    if( level == 1 )
        {
        total_symbols  = 0;
        total_matrices = 0;
        cout << "\n-------------------------------------------------------------\n";  
        cout << "Beginning feature extraction at "<< dir_name << endl << endl;
        }

    // Recursively invoke this procedure on all subtrees.  For this to work,
    // the directory search must be non-recursive.
  
    Directory tree( dir_name, D_Dirs | D_NonRec );
    while( tree.Next() ) 
        {
        tab( level );
        cout << "Directory " << tree.FullPath() << ":" << endl;
        ComputeFeatureMatrix( tree.FullPath( buff ), level + 1 );
        }

    // Process all files in this directory.  No recursion!

    tree.Open( dir_name, D_Files | D_NonRec );
  
    List feature_stats;
    while( tree.Next() ) 
        {
        if( tree == "FeatureMatrix" ) continue;
        tab( level );
        cout << "file "<< tree.ObjName() << ": ";
        FeatureStatsVector *fsv = new FeatureStatsVector;
        fsv->ProcessTrainingFile( tree.FullPath() );
        feature_stats << fsv;
        }
  
    if( feature_stats.NumEntries() > 0 ) 
        {
        WriteFeatureMatrix( dir_name, feature_stats, level );
        }
  
    if( level == 1 )
        {
        cout << "End of feature extraction at " << dir_name << endl;
        cout << "Encoded a total of " << total_symbols << " symbols in " 
             << total_matrices << " FeatureMatrix file(s)." << endl;
        cout << "-------------------------------------------------------------";
        cout << endl << endl;
        }
    }

/***************************************************************************
* Main program for PostTrain.  Handle command-line arguments, then
* start recursive computation of feature matrices in all sub-directories.
****************************************************************************/
int main( int argc, char *argv[] )
    {
    // Make the command line arguments available to any module
    // that wants to interrogate them.
    CommandLine args( argc, argv );
    const char *dir_name = ".";

    // Handle the "version" flag.
    if( args.GetSwitch( "v" ) )
        {
        cout << "Version of Software       = " << SoftwareVersion << endl;
        cout << "Version of Training Files = " << TD_Version      << endl;
        cout << "Version of Feature Matrix = " << FM_Version      << endl;
        exit(0);
        }

    // Handle the "file" flag.
    if( args.GetSwitch( "d" ) )
        {
        dir_name = args.GetArg( "d" );
        }

    // Handle the "help" flag.
    if( args.GetSwitch( "h" ) )
        {
        cout << endl;
        cout << "Usage: PostTrain { -[dhv] [option] }*" << endl         << endl;
        cout << "  -d <dname>  Read training data from this directory." << endl;
        cout << "  -h          Print these instructions and quit."      << endl;
        cout << "  -v          Print the version number and quit."      << endl;
        cout << endl;
        exit(0);
        }

    ComputeFeatureMatrix( dir_name );
    return 0;
    }








