/***************************************************************************
* Train.C                                                                  *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/18/99  arvo    Completely rewritten.                            *
*       08/15/99  arvo    Removed fixed-length arrays.                     *
*       08/06/99  arvo    Restructuring and cleanup.                       *
*       12/21/97  arvo    major cleanup.                                   *
*       12/01/96  guo     Initial codiing.                                 *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <string.h>
#include <util/File.h>
#include <util/Args.h>
#include <util/Char.h>
#include <rec/Font.h>
#include <rec/Train.h>
#include <rec/Features.h>

// Allocate and initialize all the static data members of the Train class.

Profile     Train::profile;
Train      *Train::train        = NULL;
Tcl_Interp *Train::interp       = NULL;
const char *Train::proc_name    = NULL;
int         Train::debug        = 0;
int         Train::do_post_proc = 0;
char        Train::data_dir[ Max_Path_Length ]; // Where the training data is located.
char        Train::font_dir[ Max_Path_Length ]; // Where the font descriptions live.
char        Train::base_dir[ Max_Path_Length ]; // The path used to invoke Train.

static int       startup_mode = interactive_mode;
static const int Buffer_Size = 4096;
static char      Buffer[ Buffer_Size ] = { NullChar };

/***************************************************************************
 * Check that the version of Tcl/Tk is recent enough, and initialize all
 * data members.  In particular, set the "base directory" to either the
 * supplied path, or to "./" -- this is where to look for the Tcl/Tk file(s).
 * The "data directory" is where to look for training files and/or feature
 * matrices.
 **************************************************************************/
Train::Train( const char *path )
    {
    int here = ( path == NULL ) || ( strlen( path ) == 0 );
    if( TCL_MAJOR_VERSION < 8 )
        {
        cerr << "Tcl/Tk version 8.0 or better is required by this application." << endl;
        cerr << "The installed version is " << TCL_MAJOR_VERSION << endl;
        exit(0);
        }
    curr_choice    = 0;
    stats_flag     = 0;
    do_post_proc   = 0;
    debug          = CommandLine::GetSwitch( "x" );
    proc_name      = NULL;
    exemplar_index = 0;
    strcpy( base_dir, here ? "./" : path );
    strcpy( data_dir, "." );
    grouper.SetRecognizer( *this );
    grouper.SetLookahead( 6 );
    }

Train::~Train()
    {
    // There is currently nothing to destroy in the Train class.
    }

/***************************************************************************
 * Start up a Tcl interpreter and pass it the name of the C++ function
 * that will initialize Tcl & Tk, create new Tcl command, and finally
 * start the interface running.  CONTROL NEVER RETURNS FROM "RUN".
 **************************************************************************/
void Train::Run( 
    int argc, 
    char *argv[], 
    const char *data_dir_name, 
    const char *font_dir_name, 
    int mode )

    {
    train = this;
    strcpy( data_dir, data_dir_name );
    strcpy( font_dir, font_dir_name );
    startup_mode = mode;
    Tk_Main( argc, argv, Train::Startup );
    // Control never returns from Tk_Main.
    assert(0);
    }

/***************************************************************************
 * This function is called by Tk_Main after creating an interpreter but
 * before starting the event loop.  This is where any special initialization
 * of the Tcl environment must be performed -- such as establishing the
 * "SendMessage" command that Tcl will use to communicate with C++.
 **************************************************************************/
int Train::Startup( Tcl_Interp *interp_ptr )
    {
    interp = interp_ptr;

    if( Tcl_Init( interp ) != TCL_OK ) 
        {
        cerr << "Warning! Tcl_Init failed!\n";
        exit(0);
        }

    if( Tk_Init( interp ) != TCL_OK ) 
        {
        cerr << "Warning! Tk_Init failed!\n";
        exit(0);
        }

    // Create only one new command for the Tcl environment.  All further
    // dispatching is done at this end.

    Tcl_CreateCommand( interp, "SendMessage", SendMessage, NULL, NULL );

    // Tell Tcl where to find the ".tcl" script to run.  Look for it in
    // the same directory that the executable was found.

    static char tcl_path[128], version[32];
    sprintf( tcl_path, "%strain.tcl", base_dir );
    sprintf( version , "%3.1f", SoftwareVersion );
    Tcl_SetVar( interp, "tcl_rcFileName" , tcl_path, 0 );
    Tcl_SetVar( interp, "software_version", version , 0 );

    // Tell Tcl which interface to bring up by initializing a
    // variable called "startup_mode" in the Tcl environment.

    char *mode = NULL;
    switch( startup_mode )
        {
        case interactive_mode: mode = "interactive_mode"; break;
        case regression_only : mode = "regression_only" ; break;
        case thumbnails_only : mode = "thumbnails_only" ; break;
        }
    assert( mode != NULL );
    Tcl_SetVar( interp, "startup_mode", mode, 0 );

    if( !nullString( font_dir ) )
        {
        //        FontTable::FT.Verbose( cerr );
        FontTable::FT.ReadDirectory( font_dir );
        }

    return TCL_OK;
    }

/***************************************************************************
 * Load all the feature matrices that are found in the tree whose root is
 * given by "path".  WHen done, tell Tcl how many symbols were defined so
 * that this information can be posted for the user.
 **************************************************************************/
void Train::Load( const char *path ) 
    {
    if( !is_Dir( path ) )
        {
        cerr << "Directory " << path << " does not exist." << endl;
        exit(0);
        }
    LoadFeatureMatrices( path );
    Tcl_Command( "SetNumSymbols", NumSymbols() );
    }

void Train::Tcl_Command( const char *cmd1, const char *cmd2, const char *cmd3 ) const
    {
    Tcl_VarEval( interp, cmd1, " ", cmd2, " ", cmd3, NULL );
    }

void Train::Tcl_Command( const char *cmd, int arg ) const
    {
    char buffer[256];
    sprintf( buffer, "%s %d", cmd, arg );
    Tcl_VarEval( interp, buffer, NULL );
    }

void Train::Tcl_Command( const char *cmd, int arg1, int arg2 ) const
    {
    char buffer[256];
    sprintf( buffer, "%s %d %d", cmd, arg1, arg2 );
    Tcl_VarEval( interp, buffer, NULL );
    }

int Train::BadArgs( int argc, char ** )
    {
    const char *name = "<Unknown>";
    if( proc_name != NULL ) name = proc_name;
    sprintf( interp->result, "%s: Wrong number of arguments (%d)", name, argc );
    return TCL_ERROR;
    }

/***************************************************************************
 * Check for strings that will cause trouble if passed to Tcl directly, such 
 * as "{" and "}".  Return the appropriately "escaped" version.
 **************************************************************************/
const char *FilterTclName( const char *name )
    {
    if( strlen( name ) == 1 ) switch( name[0] )
        {
        case '%'  :  return "%%"  ;
        case '$'  :  return "\\$" ;
        case '['  :  return "\\[" ;
        case ']'  :  return "\\]" ;
        case '{'  :  return "\\{" ;
        case '}'  :  return "\\}" ;
        case '"'  :  return "\\\"";
        case '\\' :  return "\\"  ;
        default   :  return name  ;
        }
    return name;
    }

/***************************************************************************
 * Generate a unique file name by appending a sufficiently large number to 
 * the name of the symbol.  Also, add a "_" prefix if the name begins with
 * a lower-case letter, to avoid problems on Windows platforms.
 **************************************************************************/
static char *GenerateUniqueName( const char *symb, char *path )
    {
    static char filename[ Max_Path_Length ];
    static char newsymb [ Max_Name_Length ];

    // See if the symbol begins with a lower-case letter.  Add a prefix
    // of "_" if it does.  This will avoid name collisions on operating
    // systems that do not support case-sensitive file names.
    if( isLcase( symb[0] ) )
         sprintf( newsymb, "_%s", symb );
    else sprintf( newsymb, "%s" , symb );

    // Put an upper limit on the search just in case the "is_File"
    // predicate is consistently failing for some reason (such as
    // the file system being full).
    for( int i = 0; i < 1000; i++ )
        {
        sprintf( filename, "%s/%s.%d", path, newsymb, i ); 
        if( !is_File( filename ) )
            return filename;
        }
    return "";
    }

/***************************************************************************
 * Write all the multilines stored in the "Samples" array to the file of the
 * given name.  Also write out a header indicating the version of the 
 * training file, and how many samples are included.
 **************************************************************************/
void Train::WriteTrainingFile( const char *name )
    { 
    char path[ Max_Path_Length ];

    // Use the user-supplied name of the symbol as the file name,
    // but tack on an extention to make it unique, and flag lower-case
    // characters.
    strcpy( path, GenerateUniqueName( name, data_dir ) );
 
    filebuf fbuff;
    fbuff.open( path, ios::out );
    ostream out( &fbuff );
  
    if( !out )
        {
        // This should never happen, but just in case...
        cerr << "ERROR: could not open file " << path 
             << " to save training data." << endl;
        }
    else
        {
        out << form( TD_VersionFormat, TD_Version )
            << num_samples << endl
            << name        << endl;

        // Write the "examplar" sample first, followed by all the rest.
        // This makes the standard sample easy to access.
        out << Samples[ exemplar_index ];
        for( int i = 0; i < num_samples; i++ ) 
            if( i != exemplar_index ) out << Samples[i];
        fbuff.close();
        do_post_proc = 1; // Run "PostTrain" when Training is done.
        }
    }

/***************************************************************************
 * Draw a stored sample in the main canvas.  This allows us to re-inspect
 * and possibly delete previously-drawn samples.
 **************************************************************************/
void Train::RedrawSample( int k ) const
    {
    if( k < 0 || num_samples == 0 ) return;
    int i = ( k < num_samples ) ? k : num_samples - 1;
    Tcl_Command( "ClearCanvas" );  
    DrawMultiline( Samples[i] );
    }   

void Train::FillMultiline( const Multiline &lines, const char *color ) const
    {
    Vec2 P;
    int len = 0;
    char col[64];
    sprintf( col, "\" %s", color );
    for( Pnode *n = lines.First(); n != NULL; n = n->Next() )
        {
        if( n->Jump() && len > 0 )
            {
            Tcl_Command( "FillPolygon \"", Buffer, col ); 
            len = 0;
            }
        assert( len < Buffer_Size - 20 );
        P = n->Coord();
        len += sprintf( Buffer + len, "%1.0f %1.0f ", P.X(), P.Y() );
        }
    if( len > 0 ) Tcl_Command( "FillPolygon \"", Buffer, col );
    }   

void Train::DrawMultiline( const Multiline &lines, const char *color ) const
    {
    Pnode *q = NULL;
    for( Pnode *p = lines.First(); p != NULL; p = p->Next() )
        {
        if( !p->Jump() && q != NULL ) DrawLine( *p, *q, color );
        q = p;
        }
    }   

void Train::DrawLine( const Vec2 &P, const Vec2 &Q, const char *color ) const
    {
    static char buff[128];
    sprintf( buff, "%d %d %d %d", (int)P.X(), (int)P.Y(), (int)Q.X(), (int)Q.Y() );
    Tcl_Command( "DrawLine", buff, color );
    }

/***************************************************************************
 *
 *
 *
 **************************************************************************/
void Train::RunRecognizer( const Multiline &lines )
    {
    static char ans[64];
    if( !lines.Empty() )
        {
        FeatureRecognizer::Recognize( lines, profile );
        if( debug ) cerr << profile;
        UpdateProfile( profile );
        symbol = profile.GetSymbol(0);

        strcpy( ans, FilterTclName( symbol->Name() ) );
        if( strlen( ans ) == 1 )
            {
            if( isUcase( ans[0] ) ) strcat( ans, " (upper case)" );
            if( isLcase( ans[0] ) ) strcat( ans, " (lower case)" );
            }
        Tcl_Command( "PostAnswer {", ans, "}" );
        Tcl_Command( "EnableYesNo" );
        if( stats_flag ) 
            { 
            curr_choice = 0; 
            UpdateStatistics( curr_choice ); 
            }

        // See if there is a stroke or outline font associated with this
        // symbol.  If not, disable the "Morph" button.
        Symbol *the_symbol = FontTable::FT.Lookup( symbol->Name(), "<font>" );
        Tcl_Command( "EnableMorph", the_symbol != NULL );

        // if( the_symbol != NULL ) cout << the_symbol->Outline();
        }
    }

/***************************************************************************
 * Repackage the data in the profile (which was returned by the recognizer) 
 * so that it can be handed off a Tcl procedure.  The Tcl procedure then 
 * updates the bar chart.
 **************************************************************************/
void Train::UpdateProfile( Profile &prof )
    {
    char buff[1024];
    char *b = buff;
    b += sprintf( b, "{" );
    float max_prob = prof.GetProb(0);
    for( int i = 0; i < prof.NumChoices(); i++ )
        {
        // Ensure that all symbols are appropriately encoded so that
        // Tcl will not get confused by characters such as "{" and "$".
        const char *str = FilterTclName( prof.GetSymbol(i)->Name() );
        float prob = prof.GetProb(i);
        b += sprintf( b, "{\"%s\" %f %f} ", str, prob / max_prob, prob );
        }
    sprintf( b, "}" );
    Tcl_Command( "UpdateProfile", buff );
    }

/***************************************************************************
 *
 *
 *
 **************************************************************************/
void Train::UpdateStatistics( int choice )
    {
    const int buff_size = 10000;
    FeatureVector FV;
    FeatureStats fs;
    char buff[ buff_size ];
    char *b = buff;
    if( choice < profile.NumChoices() )
        {
        Symbol *lett = profile.GetSymbol( choice );
        float   prob = profile.GetProb( choice );
        const FeatureStatsVector &fsv = GetStats( lett );
        b += sprintf( b, "{%s %s %g} {", lett->Name(), lett->Qual(), prob );
        ExtractFeatures( Strokes, FV );
        for( int i = 0; i < FV.Size(); i++ )
            {
            b += sprintf( b, "{%g \"%s\"} ", FV.Feature(i), FV.Comment(i) );
            }
        b += sprintf( b, "} {" );
        for( int j = 0; j < FV.Size(); j++ )
            {
            fs = fsv.Stats(j);
            b += sprintf( b, "{%g %g %g %g} ", fs.Average, fs.StndDev, fs.Min, fs.Max );
            }
        b += sprintf( b, "} " );
        }
    assert( b - buff < buff_size );
    Tcl_Command( "PostFeatureStats", buff );
    }

/***************************************************************************
 * This is the only C++ procedure that is invoked from Tcl.  From here
 * the appropriate procedures are called by dispatching on the first 
 * argument, which is assumed to be the name of the C++ function to call.
 * The remaining arguments (if there are any) are then passed to that
 * C++ function via "argc" and "argv".
 **************************************************************************/
int Train::SendMessage( ClientData, Tcl_Interp *, int argc, char *argv[] )
    {
    if( argc < 2 ) 
        {
        cerr << "ERROR: no args passed to SendMessage." << endl;
        return TCL_ERROR;
        }
    // Save a pointer ot the name of the procedure to dispatch on (which
    // is actually the 1'st argument to "SendMessage").  This is used
    // only for error reporting.
    proc_name = argv[1];
    if( debug > 2 )
        {
        cerr << "Message: ";
        for( int i = 1; i < argc; i++ ) cerr << argv[i] << " ";
        cerr << endl;
        }
    return train->Dispatch( argc, argv );
    }

/***************************************************************************
 * Run the recognition procedure on all of the training data that was used
 * to create the feature matrices.  Keep statistics about how well the
 * recognizer performs on this data, and report the results to the user.
 **************************************************************************/
void Train::RunRegressionTest( const char *frame )
    {
    Multiline lines;
    int   num_files  = 0;
    int   num_errors = 0;
    int   num_tests  = 0;
    int   total_symb = NumSymbols();
    int   index      = 0;
    int   num        = 0;
    float ave        = 0.0;
    float max        = 0.0;
    char  sym[64];
    char  frac[64];
    Directory file( data_dir, D_Files );
    Profile prof;

    Tcl_Command( "RegressionTestBegin", frame, "; update" );
    if( debug != 0 ) cerr << "Running Regression Test..." << endl;
    strcpy( sym, "" );

    // Loop through all elements of the directory looking for training
    // files.  Open each training file, read the samples, and invoke the
    // recognizer on each sample.  Keep track of how well the recognizer
    // does.
    while( file.Next() )
        {
        double uncert = 0.0;
        float  version = 0;
        int    num_samples;
        char   symb_name[64];
        char   in_buffer[256];

        if( file == "FeatureMatrix" ) continue;
        num_files++;
        filebuf fbuff;
        fbuff.open( file.FullPath(), ios::in );
        istream in( &fbuff );
        if( !in ) continue;
        in.getline( in_buffer, sizeof(in_buffer) );
        int count = sscanf( in_buffer, TD_VersionFormat, &version );
      
        if( count != 1 ) 
            {
            cerr << file.FullPath() << " Invalid format" << endl;
            continue;
            }
        if( version != TD_Version ) 
            {
            cerr << file.FullPath() << " Wrong version of training file" << endl;
            continue;
            }

        in >> num_samples;
        in >> symb_name;
        sprintf( frac, "%f", float( ++index ) / total_symb );
        Tcl_Command( "RegressionTestProgress", symb_name, frac );
        if( debug != 0 ) cerr << "  Testing " << symb_name << endl;

        // Now read all the samples of this symbol.
        for( int samp = 0; samp < num_samples; samp++ )
            {
            in >> lines;
            if( lines.Empty() ) continue;
            FeatureRecognizer::Recognize( lines, prof );
            Symbol *symbol = prof.BestSymbol();
            if( !Match( symbol->Name(), symb_name ) )
                {
                Tcl_Command( "RegressionTestError", symb_name, symbol->Name() );
                num_errors++;
                }
            num_tests++;
            uncert = prof.Uncertainty( symb_name );
            ave += uncert;
            if( uncert > max ) 
                {
                max = uncert;
                num = samp;
                strcpy( sym, symb_name );
                }
            }
        fbuff.close();
        }
    ave /= num_tests;

    // Send the results to Tcl.
    float rate = (100.0 * ( num_tests - num_errors )) / num_tests;
    Tcl_Command( 
        "RegressionTestResults", 
        form( "{{nerr %d} {ntst %d} {rate %g}", num_errors, num_tests, rate ),
        form( " {ave %g} {max %g %s %d}}", ave, max, FilterTclName( sym ), num )
        );
    }

/***************************************************************************
 * Create a panel that shows thumbnail images of all the symbols that are
 * currently recognized by the recognizer.  This pannel will also show
 * animations of the symbols being drawn.
 **************************************************************************/
void Train::ShowThumbnails( const char *top ) const
    {
    static int N = 28; // Resolution of the thumbnail images.
    char symb_name[64];
    char buf[64];
    char fname[ Max_Name_Length ];
    char pname[ Max_Path_Length ];
    int  first = 1;
    Multiline lines;
    Directory file( data_dir, D_Files | D_Dirs | D_PreOrder );
    SortedList list;
    Tcl_Command( "Thumb_Begin", top );

    // Loop through all elements of the directory looking for training files.
    for(;;)
        {
        if( file.is_Dir() || file.End() )
            {
            if( list.Size() > 0 )
                {
                for( int i = 0; i < list.Size(); i++ )
                    {
                    if( GetExemplar( list(i), lines, symb_name ) )
                        {
                        sprintf( buf, "Thumb_AddImage %d", N );
                        Tcl_Command( buf, symb_name, list(i) );
                        DrawThumbnail( N, lines );
                        }
                    }
                list.Clear();
                }
            if( file.End() ) break;
            strcpy( fname, file.ObjName () );
            strcpy( pname, file.FullPath() );
            first = 1;
            }
        else if( file.is_File() && file != "FeatureMatrix" )
            {
            if( first ) Tcl_Command( "Thumb_Section", fname, pname );
            list << file.FullPath();
            first = 0;
            }
        file.Next();
        }
    Tcl_Command( "Thumb_End" );
    }

/***************************************************************************
 * Draw a thumbnail image of the multiline into the small canvas.
 **************************************************************************/
void Train::DrawThumbnail( int size, const Multiline &lines ) const
    {
    char coords[64];
    Multiline P( lines );
    P -= P.Box().Center();
    float height = P.Box().Height();
    float width  = P.Box().Width ();
    float max    = Max( height, width );
    float scale  = size / max;
    P *= 0.85 * scale;
    P += ( 0.5 + ( size + 6 ) / 2.0 ) * Vec2(1,1);
    Pnode *p = P.First();
    for( Pnode *q = p->Next(); q != NULL; q = q->Next() )
        {
        if( !q->Jump() ) 
            {
            sprintf( 
                coords, 
                "%d %d %d %d", 
                (int)p->X(), (int)p->Y(), (int)q->X(), (int)q->Y() 
                );
            Tcl_Command( "Thumb_Line", coords );
            }
        p = q;
        }
    }

int Train::ShowActualChar( const Multiline &strokes, Symbol *symb ) const
    {
    char coords[256];
    if( !strokes.Empty() )
        {
        Box2d box = strokes.Box();
        if( symb != NULL )
            {
            Trans2 trans = Box_to_Box( symb->Stroke().Box(), box );
            FillMultiline( trans * symb->Outline(), "gray90" );
            }
        sprintf( coords, "%g %g %g %g", box.Xmin(), box.Ymin(), box.Xmax(), box.Ymax() );
        Tcl_Command( "DrawGroupBox", coords, "orange" );
        Tcl_Command( "update" );
        return 1;
        }
    return 0;
    }
