/*****************************************************************************
** 
**   MODULE NAME  Triangulate.C
**
**   DESCRIPTION
**
**
**   HISTORY
**      Name    Date       Description
**
**      novins  
**
**
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stream.h>
#include <assert.h>
#include <util/Multiline.h>
#include <util/Delaunay.h>
#include <util/quadedge.h>

static Multiline S;
static Multiline O;

extern void Open( const char *thing );
extern void Close();
extern FILE  *file_out;

#undef X
#undef Y
#undef Z

void identify_edges( QuadEdge *q, void *data )
    {
    // An iterator for all quadedges.  Determines if each edge is internal 
    // or external depending on an inside/outside test of the multiline 
    // passed as data.  Sets the Etype field of the quadedge.

    // Identify vertices of the edge
    Multiline *mline = (Multiline *) data;
    Point2d a,b;
    a = q->edges()[0].Org2d();
    b = q->edges()[2].Org2d();
    
    // Find the midpoint of the edge
    Vec2 mid;
    mid.X() = .5 * (a[0] + b[0]);
    mid.Y() = .5 * (a[1] + b[1]);
    
    // Find a unit vector perpendicular to the edge
    Vec2 perp;
    perp.X() = b[1] - a[1];
    perp.Y() = a[0] - b[0];
    Normalize(perp);
    
    // Take a small step on either side of the midpoint.  If both sides
    // are inside the polygon, then this vertex is internal.
    perp *= .0001;
    if (mline->Inside( mid+perp) && mline->Inside( mid-perp))
        {
        q->Etype( e_internal);
        }
    else
        {
        q->Etype( e_external);
        }
    }


face_type facetype( Edge *e )
    {
    Edge *min, *cur;

    min  = e->invRot();
    cur  = e->Lnext()->invRot();
    if (cur->QEdge() < min->QEdge() ) min = cur;
    cur  = e->Lnext()->Lnext()->invRot();
    if (cur->QEdge() < min->QEdge() ) min = cur;
    return( min->Ftype() );
    }

void facetype( Edge *e, face_type t )
    {
    Edge *min, *cur;

    min  = e->invRot();
    cur  = e->Lnext()->invRot();
    if (cur->QEdge() < min->QEdge() ) min = cur;
    cur  = e->Lnext()->Lnext()->invRot();
    if (cur->QEdge() < min->QEdge() ) min = cur;
    min->Ftype(t);
    }

void identify_faces( Edge *e, void *)
    {
    facetype(e, f_none);
    }

Vec2 find_midpoint( Edge *e )
    {
    Vec2 v;
    Point2d a = e->Org2d ();
    Point2d b = e->Dest2d();
    v.X() = (a[0] + b[0])/2.0;
    v.Y() = (a[1] + b[1])/2.0;
    return(v);
    }


float SearchSleeves( Edge *e, Vec2 o, float distsofar, float value, float &targvalue, Multiline *M, int depth)
    {
    if (depth > 1000)
        {
        //fprintf(stderr,"SearchSleeves recursion error\n");
        return(-1);
        }

    facetype( e, f_marked );

    // If opposite vertex is not same value, we're done.
    if (e->Lnext()->Dest()->Value() != value )
        {
        Vec2 target = Vec2( e->Lnext()->Dest2d()[0],
                            e->Lnext()->Dest2d()[1]);
        float dfromtarget = Len( target-o);
        float alpha = distsofar/(distsofar + dfromtarget);
        targvalue = e->Lnext()->Dest()->Value();

        TimedVec2 to_add(o);
        to_add.SetTime((long)
                       ((alpha * targvalue + (1.0-alpha) * value) * 1000.0));
        *M << to_add;

        return( dfromtarget );
        }

    // Try the first unconstrained vertex we see
    Edge *temp;
    temp = e->Lnext();
    while ( temp->QEdge()->isConstrained() )
        {
        temp = temp->Lnext();
        }

    if (temp == e)
        return(-1.0);

    Vec2 m = find_midpoint( temp );
    distsofar += Len( m-o );
    float dfromtarget = SearchSleeves( temp->Sym(),m,distsofar,value,targvalue,M, depth+1 );
    if (dfromtarget < 0.0)
        {
        temp = temp->Lnext();
        if (temp != e && !(temp->QEdge()->isConstrained() ))
            {
            dfromtarget = SearchSleeves( temp->Sym(),m,distsofar,value,targvalue,M, depth+1 );
            }
        }

    if (dfromtarget < 0.0) return(dfromtarget);
    float alpha = distsofar/(distsofar + dfromtarget);
    float new_t = alpha * targvalue + (1.0-alpha) * value;

#ifdef UNDEF
    // If more than one path through the same vertex was made, set t
    // to the *largest* value to ensure a smooth morph
    int i = L->n;
    for (int j=0; j < i; j++)
        {
        if (
            ((L->x)[j] == o.X() ) &&
            ((L->y)[j] == o.Y() ) &&
            ((L->t)[j] >  new_t )
            )
            new_t = (L->t)[j];
        }
#endif

    TimedVec2 to_add(o);
    to_add.SetTime( (long)(new_t * 1000.0) );
    *M << to_add;
    return( dfromtarget + Len(m-o) );
    }


                            

void make_limbs( Edge *e, void *v )
    {
    Multiline *M = (Multiline *) v;

    // Reject faces that have already been visited
    if ( facetype( e ) == f_marked )
        return;
    
    facetype( e->Rot(), f_marked );

    // Reject faces that don't have two constraint edges
    int countconstrained = 0;
    Edge *temp;
    temp = e->Rot();
    if (temp->QEdge()->isConstrained()) countconstrained++;
    temp = temp->Lnext();
    if (temp->QEdge()->isConstrained()) countconstrained++;
    temp = temp->Lnext();
    if (temp->QEdge()->isConstrained()) countconstrained++;
    temp = temp->Lnext();

    if ( countconstrained != 2 )
        return;

    // Find the unconstrained edge.  
    // Is it internal?
    // Does it join two vertices with the same value?

    Edge *unconstrained = e->Rot();
    do {
        unconstrained = unconstrained->Lnext();
    }
    while ( unconstrained->QEdge()->isConstrained() );

    if (unconstrained->QEdge()->Etype() != e_internal )
        return;

    if (unconstrained->Org()->Value() == unconstrained->Dest()->Value())
        {
        Vec2 m = find_midpoint( unconstrained );
        Vec2 o = Vec2( unconstrained->Lnext()->Lnext()->Org2d()[0],
                       unconstrained->Lnext()->Lnext()->Org2d()[1]);
        float distsofar = Len( m-o );

        float value = unconstrained->Org()->Value();

        float targvalue=0.0;
        if ( SearchSleeves( unconstrained->Sym(),m,distsofar,value,targvalue, M, 0 ) < 0.0 )
            {
            //fprintf(stderr,"Searchsleeves error.\n");
            };
        }
    }

void make_limbs2( QuadEdge *q, void *data )
    {
    if ( q->isConstrained() ) 
        return;

    if ( q->Etype() != e_internal ) 
        return;

    if (q->edges()[0].Org()->Value() != q->edges()[2].Org()->Value())
        return;

    Multiline *M = (Multiline *) data;
    Point2d a,b;
    a = q->edges()[0].Org2d();
    b = q->edges()[2].Org2d();
    
    // Find the midpoint of the edge
    TimedVec2 mid;
    mid.X() = .5 * (a[0] + b[0]);
    mid.Y() = .5 * (a[1] + b[1]);
    float Dsrc = Len(S.Closest(mid) - mid);
    float Ddst = Len(O.Closest(mid) - mid);
    mid.SetTime( (long) (1000.0 * Dsrc/(Dsrc+Ddst)) );
    
    *M << mid;
    }

// Inserts a triangle with vertex intensities into the mesh section of 
// the ".fdf" file.
static void PutTriangle( Vec2 A, float ia, Vec2 B, float ib, Vec2 C, float ic )
    {
    fprintf( file_out, "    " );
    fprintf( file_out, "%8.5f %8.5f %8.5f ", A.X(), A.Y(), ia );
    fprintf( file_out, "%8.5f %8.5f %8.5f ", B.X(), B.Y(), ib );
    fprintf( file_out, "%8.5f %8.5f %8.5f ", C.X(), C.Y(), ic );
    fprintf( file_out, "\n" );
    }


void output_tris( Edge *e, void * )
    {
    Edge *temp;
    int  internal_edge_count = 0;

    temp = e->Rot();
    internal_edge_count += (temp->QEdge()->Etype() == e_internal ? 1 : 0);
    Vec2 A( 
           temp->Org2d()[0],
           temp->Org2d()[1]
           );
    float ia = temp->Org()->Value();

    temp = temp->Lnext();
    internal_edge_count += (temp->QEdge()->Etype() == e_internal ? 1 : 0);
    Vec2 B(
           temp->Org2d()[0],
           temp->Org2d()[1]
           );
    float ib = temp->Org()->Value();

    temp = temp->Lnext();
    internal_edge_count += (temp->QEdge()->Etype() == e_internal ? 1 : 0);
    Vec2 C(
           temp->Org2d()[0],
           temp->Org2d()[1]
           );
    float ic = temp->Org()->Value();

    // Only output triangles with at least one internal edge
    if (internal_edge_count>0)
        PutTriangle( A, ia, B, ib, C, ic );
    }


void GenerateTriangles( const Multiline &skeleton, const Multiline &outline )
    {
    Open( "mesh" );

    // Set O equal to the outline multiline -- with internal edges removed
    O.Clear();
    Pnode *p = outline.First();
    if (p->Jump() ) O << Multiline::Jump;
    O << *p;
    Pnode *q;
    for(q = p->Next(); q != NULL; q = q->Next() )
        {
        if( q->Jump() ) 
            {
            O << Multiline::Jump;
            }
        else
            {
            Vec2 mid; // Midpoint of the edge
            mid.X() = .5 * (p->X() + q->X());
            mid.Y() = .5 * (p->Y() + q->Y());
            
            Vec2 perp; // Unit vector perpendicular to edge
            perp.X() = q->Y() - p->Y();
            perp.Y() = p->X() - q->X();
            Normalize(perp);
            
            const float stepsize = .0001;
            perp *= stepsize;
            
            if (outline.Inside( mid+perp) && outline.Inside( mid-perp))
                {
                O << Multiline::Jump;
                }
            }
        O << *q;
        p = q;
        }

    // Set S equal to the skeleton multiline -- with new points added at T
    // junctions on the skeleton
    S.Clear();
    p = skeleton.First();
    if (p->Jump() ) S << Multiline::Jump;
    S << *p;
    for( q = p->Next(); q != NULL; q = q->Next() )
        {
        if( q->Jump() ) 
            {
            S << Multiline::Jump;
            }
        else
            {
            // check if this line segment needs to be split
            for (Pnode *r = skeleton.First(); r != NULL; r = r->Next())
                {
                if (r == p || r == q ) continue;
                Vec2 a(*q-*p);
                Vec2 b(*r-*p);
                float x = (a*b)/(a*a);
                if (x < 1.0 && x > 0.0) 
                    {
                    if (Len( x*a-b ) < .01)
                        {
                        S << *r;
                        // Can't add more than one point to this segment
                        // without sorting them.  To avoid the trouble,
                        // just adding at most one point to a segment now.
                        break;
                        }
                    }
                }
            }
        S << *q;
        p = q;
        }

    // Need a bounding box to make the CDT code happy.  Use a box that 
    // is 5% bigger than the "true" bounding box of the multiline, to be safe.
    Box2d b( O.TrueBox() );
    Vec2 grow( b.Width()*.05, b.Height()*.05 );
    b = Box2d( b.Min()-grow, b.Max()+grow);

    DelaunayMesh  mesh(b);

    // Add the initial multiline.  Set field values to 1 indicating that
    // these edges are on the boundary of the shape
    mesh.Add( O, 1.0 );
    mesh.Add( S, 0.0 );

    // Step through each edge, identifying internal and external edges
    ((Mesh *) mesh.GetMesh())->TraverseEdges( identify_edges, (void *) &O );

    // Add new vertices at midpoints of internal edges that connect outline
    // to outline or stroke to stroke.  
    Multiline M;
#ifdef NOTDEF
    ((Mesh *) mesh.GetMesh())->TraverseTris( make_limbs, (void *) &M );
#endif

    ((Mesh *) mesh.GetMesh())->TraverseEdges( make_limbs2, (void *) &M );

    for( q = M.First(); q != NULL; q = q->Next() )
        {
        ((Mesh *) mesh.GetMesh())->InsertValue( 
                                               Point2d(q->X(),q->Y()), 
                                               (float) q->Time()/1000.0);
        }


    // Output triangles that are internal to the outline
    ((Mesh *) mesh.GetMesh())->TraverseEdges( identify_edges, (void *) &O );
    ((Mesh *) mesh.GetMesh())->TraverseTris( output_tris, (void *) NULL );


    Close();
    }




