/***************************************************************************
* Command.C                                                                *
*                                                                          *
* This module defines all the functions that are called from Tcl/Tk via    *
* the SendMessage and Dispatch mechanisms.                                 *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/18/99  arvo    Split off from Train.C                           *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <assert.h>
#include <string.h>
#include <util/Char.h>
#include <util/Timer.h>
#include <util/Trans2.h>
#include <rec/Font.h>
#include <rec/Train.h>
#include <rec/LineMorph.h>
#include <rec/BasicGrouper.h>

static Timer clock;

int Train::Dispatch( int argc, char *argv[] )
    {
    // All C++ functions that are invoked from within Tcl via the 
    // "SendMessage" command must be listed here.
    DispatchHere( draw_to       );
    DispatchHere( move_to       );
    DispatchHere( end_line      );
    DispatchHere( timed_out     );
    DispatchHere( clear         );
    DispatchHere( post_proc     );
    DispatchHere( yes_button    );
    DispatchHere( no_button     );
    DispatchHere( set_mode      );
    DispatchHere( load_data     );
    DispatchHere( quit          );
    DispatchHere( show_profile  );
    DispatchHere( hide_profile  );
    DispatchHere( new_sample    );
    DispatchHere( erase_sample  );
    DispatchHere( save_samples  );
    DispatchHere( make_exemplar );
    DispatchHere( print_lines   );
    DispatchHere( show_stats    );
    DispatchHere( change_stats  );
    DispatchHere( run_reg_test  );
    DispatchHere( next_symbol   );
    DispatchHere( prev_symbol   );
    DispatchHere( show_thumb    );
    DispatchHere( thumb_anim    );
    DispatchHere( morph         );
    DispatchHere( group         );
    cerr << "Unrecognized command: " << argv[1] << endl;
    return TCL_OK;
    }

DispatchDecl( Train::move_to )
    {
    if( argc != 2 ) return BadArgs( argc, argv );
    int x = atoi( argv[0] );
    int y = atoi( argv[1] );
    int u;
    // If this is the first point of the strokes, ensure that
    // it is given a time stamp of zero.
    if( Strokes.Empty() ) 
        { 
        u = 0; 
        clock.Start(); 
        }
    else u = clock.RealMilliseconds();
    Strokes << TimedVec2( x, y, u );
    return TCL_OK;
    }

DispatchDecl( Train::draw_to )
    {
    if( argc != 2 ) return BadArgs( argc, argv );
    int x = atoi( argv[0] );
    int y = atoi( argv[1] );
    int u = (int)clock.RealMilliseconds();
    // Add the new sample points, provided they are unique!
    // That is, be careful to skip repreated points.
    TimedVec2 P( x, y, u );
    if( Strokes.Empty() || SupNorm( P - Strokes.End() ) > 1.0 ) Strokes << P;
    return TCL_OK;
    }

// End of one stroke.
DispatchDecl( Train::end_line )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( mode == training_mode )
        {
        // If this is the end of the first stroke in training mode, then regard
        // this as a new sample (which may still be incomplete).  Bump the number
        // of samples and update the counter on the interface.
        if( Strokes.Comps() == 1 )
            {
            curr_sample = num_samples;
            num_samples++;
            Tcl_Command( "SetNumSamples", num_samples );
            Tcl_Command( "EnableNewErase 1" );
            Tcl_Command( "EnableExemplar", curr_sample != exemplar_index );
            }
        // At the end of each stroke, update the current sample.  This way, 
        // the Samples array will always be current.
        Samples[ curr_sample ] = Strokes;
        }
    if( mode == grouping_mode )
        {
        Multiline character;
        grouper << Strokes;
        Strokes.Clear();
        Symbol *symb = grouper.GetCharacter( character );
        ShowActualChar( character, symb );
        }
    Strokes << Multiline::Jump;
    return TCL_OK;
    }

// Quit and call the PostTrain function if symbols have been added 
// or updated. (The PostTrain program updates the FeatureMatrix files).
DispatchDecl( Train::quit )
    {
    if( argc > 1 ) return BadArgs( argc, argv );
    // If quit is called with "force", then quit immediately
    // without checking.
    if( argc == 1 && Match( argv[0], "force" ) )
        {
        Tcl_Command( "exit" );
        exit(0);
        }

    // Under normal circumstances, run the post-processing
    // tool if training files have been added or modified.
    if( do_post_proc ) post_proc( argc, argv );

    // If there are unsaved samples, be sure the user really wants
    // to quit.
    if( mode == training_mode && num_samples > 0 ) 
        {
        Tcl_Command( "ConfirmQuit" );
        }
    else
        {
        Tcl_Command( "exit" );
        exit(0);
        }
    return TCL_OK;
    }

DispatchDecl( Train::post_proc )
    {
    if( argc > 1 ) return BadArgs( argc, argv );
    char command[ Max_Path_Length ];
    sprintf( command, "./PostTrain -d %s", data_dir );
    system( command );
    if( argc == 1 && Match( argv[0], "reload" ) )
        {
        LoadFeatureMatrices( data_dir );
        Tcl_Command( "SetNumSymbols", NumSymbols() );
        }
    do_post_proc = 0;
    return TCL_OK;
    }

DispatchDecl( Train::set_mode )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    if( Match( argv[0], "rec"   ) ) mode = recognition_mode; else
    if( Match( argv[0], "group" ) ) mode = grouping_mode;    else
    if( Match( argv[0], "train" ) ) mode = training_mode;    else
    cerr << "Undefined mode: " << argv[0] << endl;
    num_samples = 0;
    grouper.Clear();
    return TCL_OK;
    }

DispatchDecl( Train::timed_out )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( mode == recognition_mode )
        {
        RunRecognizer( Strokes );
        }
    return TCL_OK;
    }

DispatchDecl( Train::yes_button )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    Strokes.Clear();
    return TCL_OK;
    }

DispatchDecl( Train::no_button )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    Strokes.Clear();
    return TCL_OK;
    }

DispatchDecl( Train::load_data )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    // Load the feature matrices from the directory given on the command line.
    Load( data_dir );
    return TCL_OK;
    }

DispatchDecl( Train::show_profile )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    UpdateProfile( profile );
    return TCL_OK;
    }

DispatchDecl( Train::hide_profile )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    return TCL_OK;
    }

DispatchDecl( Train::new_sample )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( mode == training_mode )
        {
        Strokes.Clear();
        curr_sample = num_samples;
        Tcl_Command( "ClearCanvas" );
        Tcl_Command( "EnableNewErase 0" );
        Tcl_Command( "EnableExemplar 0" );
        Tcl_Command( "SetSampleLabel", curr_sample );
        }
    return TCL_OK;
    }

DispatchDecl( Train::erase_sample )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( mode == training_mode && !Strokes.Empty() && num_samples > 0 )
        {
        if( curr_sample < num_samples ) Tcl_Command( "SetNumSamples", --num_samples );
        if( num_samples == 0 ) // The user just erased the last of the samples.
            {
            Strokes.Clear();
            exemplar_index = 0;
            Tcl_Command( "ClearCanvas" );
            Tcl_Command( "SetSampleLabel", 0 );
            Tcl_Command( "EnableNewErase 0" );
            }
        else // There is at least one sample remaining.
            {
            if( curr_sample == exemplar_index ) exemplar_index = 0;
            if( exemplar_index > curr_sample  ) exemplar_index--;
            for( int i = curr_sample; i < num_samples; i++ ) Samples[i] = Samples[i+1];
            if( curr_sample == num_samples ) 
                {
                Tcl_Command( "ClearCanvas" );
                Tcl_Command( "EnableNewErase 0" );
                Strokes.Clear();
                }
            else
                {
                RedrawSample( curr_sample );
                Tcl_Command( "EnableNewErase 1" );
                Strokes = Samples[ curr_sample ];
                }
            Tcl_Command( "SetSampleLabel", curr_sample, curr_sample == exemplar_index );
            }
        }
    return TCL_OK;
    }

DispatchDecl( Train::save_samples )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    WriteTrainingFile( argv[0] );
    num_samples = 0;
    curr_sample = 0;
    Tcl_Command( "SetNumSamples 0" );
    Tcl_Command( "EnableNewErase 0" );
    Tcl_Command( "SetSampleLabel", 0, 1 );
    return TCL_OK;
    }

DispatchDecl( Train::clear )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    Strokes.Clear();
    return TCL_OK;
    }

// Make the current (i.e. the last) sample the first in the list.
DispatchDecl( Train::make_exemplar )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( num_samples > 0 ) exemplar_index = curr_sample;
    Tcl_Command( "EnableExemplar 0" );
    Tcl_Command( "SetSampleLabel", curr_sample, 1 );
    return TCL_OK;
    }

DispatchDecl( Train::print_lines )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    cout << Strokes;
    return TCL_OK;
    }

DispatchDecl( Train::show_stats )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    if( Match( argv[0], "on"  ) ) stats_flag = 1;
    if( Match( argv[0], "off" ) ) stats_flag = 0;
    if( stats_flag ) Tcl_Command( "ShowFeatureStats", NumFeatures() );
    return TCL_OK;
    }

DispatchDecl( Train::run_reg_test )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    const char *frame = argv[0];
    RunRegressionTest( frame );
    return TCL_OK;
    }

DispatchDecl( Train::change_stats )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    int delta = atoi( argv[0] );
    if( delta == 0 )
        {
        curr_choice = 0;
        }
    else
        {
        int n = profile.NumChoices();
        if( n > 0 && delta != 0 )
            {
            curr_choice += delta;
            while( curr_choice <  0 ) curr_choice += n;
            while( curr_choice >= n ) curr_choice -= n;
            }
        }
    UpdateStatistics( curr_choice );
    return TCL_OK;
    }

DispatchDecl( Train::next_symbol )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( curr_sample < num_samples - 1 )
        {
        RedrawSample( ++curr_sample );
        int exemp = ( curr_sample == exemplar_index );
        Tcl_Command( "EnableNewErase" );
        Tcl_Command( "EnableExemplar", !exemp );
        Tcl_Command( "SetSampleLabel", curr_sample, exemp );
        Strokes = Samples[ curr_sample ];
        }
    return TCL_OK;
    }

DispatchDecl( Train::prev_symbol )
    {
    if( argc != 0 ) return BadArgs( argc, argv );
    if( curr_sample > 0 )
        {
        RedrawSample( --curr_sample );
        int exemp = ( curr_sample == exemplar_index );
        Tcl_Command( "EnableNewErase" );
        Tcl_Command( "EnableExemplar", !exemp );
        Tcl_Command( "SetSampleLabel", curr_sample, exemp );
        Strokes = Samples[ curr_sample ];
        }
    return TCL_OK;
    }

DispatchDecl( Train::show_thumb )
    {
    if( argc != 1 ) return BadArgs( argc, argv );
    const char *frame = argv[0];
    ShowThumbnails( frame );
    return TCL_OK;
    }

DispatchDecl( Train::thumb_anim )
    {
    int  delta  = 25;
    int  offset = 80;
    int  time   = offset;
    char coords[64];
    static Multiline lines;
    if( argc == 0 )
        {
        Pnode *p = lines.First();
        for( Pnode *q = p->Next(); q != NULL; q = q->Next() )
            {
            if( q->Jump() ) 
                {
                // Add a delay between strokes for samples without
                // actual timing info.
                time += 150; 
                }
            else
                {
                if( p->Time() == 0 ) time += delta;
                else time = offset + (int)( 1.1 * p->Time() );
                sprintf( coords, "%d %d %d %d %d", 
                    time,                    
                    (int)p->X(), (int)p->Y(), 
                    (int)q->X(), (int)q->Y() 
                    );
                Tcl_Command( "Thumb_Anim_Line", coords );
                }
            p = q;
            }
        return TCL_OK;
        }
    if( argc != 2 ) return BadArgs( argc, argv );
    const char *symb = argv[0];
    const char *file = argv[1];
    if( GetExemplar( file, lines ) )
        {
        lines -= lines.Box().Center();
        lines *= 140.0 / lines.Box().Diam();
        lines += 0.5 * Vec2(150,150);
        Tcl_Command( "Thumb_Animate", symb, file );
        }
    return TCL_OK;
    }

void Train::DrawTriangle( const Pnode *A, const Pnode *B, const Pnode *C, float t, const char *color ) const
    {
    long isoval = (long) (t*1000.0);

    // Set local variables to the "intensity" value at each vertex
    long a = A->Time();
    long b = B->Time();
    long c = C->Time();
    int  a_low = (a < isoval ? 1 : 0 );
    int  b_low = (b < isoval ? 1 : 0 );
    int  c_low = (c < isoval ? 1 : 0 );

    // Reject triangles that are outside the set
    if (a_low + b_low + c_low == 0) return; 
    
    char coords[1024];
    char col[1024];
    sprintf(col, "\" %s", color );

    Vec2 pts[6]; // Shouldn't need > 4; be safe since error check happens later
    int  npts=0;
    if ( a < isoval)
        {
        pts[npts] = *A;
        npts++;
        }

    if ( a_low + b_low == 1 )
        {
        float alpha = (float) (isoval-a)/ (float) (b-a);
        alpha = 1.0-alpha;
        pts[npts] = alpha * *A + (1.0-alpha) * *B;
        npts++;
        }

    if ( b < isoval)
        {
        pts[npts] = *B;
        npts++;
        }

    if ( b_low + c_low == 1 )
        {
        float alpha = (float) (isoval-c)/ (float) (b-c);
        alpha = 1.0-alpha;
        pts[npts] = alpha * *C + (1.0-alpha) * *B;
        npts++;
        }

    if ( c < isoval)
        {
        pts[npts] = *C;
        npts++;
        }

    if ( a_low + c_low == 1 )
        {
        float alpha = (float) (isoval-a)/(float) (c-a);
        alpha = 1.0-alpha;
        pts[npts] = alpha * *A + (1.0-alpha) * *C;
        npts++;
        }

    switch (npts)
        {
        case 3:
            sprintf( 
                    coords, 
                    "%d %d %d %d %d %d", 
                    (int)pts[0].X(), (int)pts[0].Y(),
                    (int)pts[1].X(), (int)pts[1].Y(),
                    (int)pts[2].X(), (int)pts[2].Y()
                );
            Train::Tcl_Command( "FillPolygon \"", coords, col );
            break;
        case 4:
            sprintf( 
                    coords, 
                    "%d %d %d %d %d %d %d %d",
                    (int)pts[0].X(), (int)pts[0].Y(),
                    (int)pts[1].X(), (int)pts[1].Y(),
                    (int)pts[2].X(), (int)pts[2].Y(),
                    (int)pts[3].X(), (int)pts[3].Y()
                    );
            Train::Tcl_Command( "FillPolygon \"", coords, col );
            break;
        default:
            fprintf(stderr,"Warning: npts=%d; should be 3 or 4.\n", npts);
            fprintf(stderr,"   A=%d B=%d C=%d isoval = %d\n",a,b,c,isoval);
            break;
        }
    
    }

void Train::DrawMesh( const Multiline &M, float t, const char *color ) const
    {
    Pnode *temp = M.First();

    while (temp != NULL )
        {
        Pnode *A = temp;
        if ( !A->Jump() )
            {
            fprintf(stderr, "Mesh polyline is not a set of triangles A\n");
            return;
            }

        temp = temp->Next();
        Pnode *B = temp;
        if ( B == NULL )
            {
            fprintf(stderr, "Mesh polyline is not a set of triangles B\n");
            return;
            }

        temp = temp->Next();
        Pnode *C = temp;
        if ( C == NULL )
            {
            fprintf(stderr, "Mesh polyline is not a set of triangles C\n");
            return;
            }

        DrawTriangle( A, B, C, t, color );

        temp = C->Next();
        }
    }


DispatchDecl( Train::morph )
    {
    int   steps = 15;
    float delay = 0.03;
    if( argc != 0 ) return BadArgs( argc, argv );
    Symbol *first = FontTable::FT.Lookup( symbol->Name(), "<font>" );
    Tcl_Command( "ClearCanvas simple" );
    DrawMultiline( Strokes );
    if( first != NULL )
        {
        Trans2 trans = Box_to_Box( first->Stroke().Box(), Strokes.Box() );
        Multiline P( Strokes );
        Multiline Q( trans * first->Stroke() );
        Multiline R;
        LineMorphSetup( P, Q );
		int i;
        for( i = 1; i <= steps; i++ )
            {
            float t = float(i) / steps;
            if( i == steps ) 
                 R = trans * first->Stroke();
            else R = Interpolate( t, P, Q );
            Timer::Sleep( delay );
            Tcl_Command( "ClearCanvas simple" );
            DrawMultiline( R );
            Tcl_Command( "update" );
            /*
            if( save ) 
                {
                char fname[128];
                sprintf( fname, "morph.%d.eps", i );
                Tcl_Command( "SavePostscript", fname );
                }
            */
            }
        Multiline M;
        M = first->Mesh();
        M *= trans.Scale();
        M += trans.Origin();
        for( i = 1; i <= steps; i++ )
            {
            float t = float(i) / steps;
            DrawMesh( M, t );
            Tcl_Command( "update" );
            //Timer::Sleep( delay );
            }

        Tcl_Command( "EnableYesNo" );
        }
    return TCL_OK;
    }


DispatchDecl( Train::group )
    {
    int force = 1;
    Multiline character;
    if( argc != 1 ) return BadArgs( argc, argv );
    if( Match( argv[0], "done" ) )
        {
        for(;;)
            {
            Symbol *symb = grouper.GetCharacter( character, force );
            if( !ShowActualChar( character, symb ) ) break;
            }
        grouper.Clear();
        }
    else if( Match( argv[0], "clear" ) )
        {
        grouper.Clear();
        Strokes.Clear();
        Tcl_Command( "ClearCanvas" );
        }
    return TCL_OK;
    }











