/***************************************************************************
* Recognizer.C                                                             *
*                                                                          *
* Implements the base class for all character recognizers.  Most of the    *
* behavior of a recognizer is defined in the corresponding header file     *
* (which defines the base class), so there is very little to implement     *
* here.                                                                    *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       10/22/99  arvo    Removed modifier "Generic" from class names.     *
*       12/19/97  arvo    Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <math.h>
#include <assert.h>
#include <stream.h>
#include <rec/Recognizer.h>

void Profile::Clear()
    { 
    num_choices = 0; 
    for( int i = 0; i < MaxProfileChoices; i++ )
        {
        S[i] = NULL;
        P[i] = 0.0;
        }
    }

Profile &Profile::operator=( const Profile &prof )
    {
    Clear();
    num_choices = prof.NumChoices();
    for( int i = 0; i < num_choices; i++ )
        {
        S[i] = prof.GetSymbol(i);
        P[i] = prof.GetProb  (i);
        }
    return *this;
    }

void Profile::Normalize()
    {
    float max = 0;
    for( int i = 0; i < num_choices; i++ ) if( P[i] > max ) max = P[i];
    if( max > 0.0 )
        for( int j = 0; j < num_choices; j++ ) P[j] /= max;
    }

void Profile::SetChoices( int num )
    {
    num_choices = num;
    if( num > MaxProfileChoices ) num_choices = MaxProfileChoices;
    if( num < 1                 ) num_choices = 1;
    }

/***************************************************************************
* As a heuristic measure of uncertainty, compute the 2-norm of all
* probabilities normalized by the probability of the most-likely choice.  
* Thus, when one or more choices have nearly as high a likelihood as the 
* most-likely choice, the uncertainty is high.
****************************************************************************/
float Profile::Uncertainty( const char *symbname ) const
    {
    if( NumChoices() < 2 ) return 0.0;
    double sum_sqr = 0.0;
    double normal  = BestProb();
    if( normal <= 0.0 ) return 0.0;
    int ord = GetOrdinal( symbname );

    // Sum the squares of all probabilities except for the
    // one corresponding to symbname.  If symbname does not
    // appear in the list, then sum all the squares.
    for( int i = 0; i < NumChoices(); i++ ) 
        if( i != ord ) sum_sqr += pow( GetProb(i), 2 );

    // Normalize according to the probability of symbname,
    // or the least-likely choice if symbname is not in the list.
    if( ord < NumChoices() ) 
         normal = GetProb( ord );
    else normal = GetProb( NumChoices() - 1 );
    assert( normal > 0.0 );

    return sqrt( sum_sqr / ( normal * normal ) );
    }

/***************************************************************************
* Look up the index of the choice that matches the given symbol name.
* If the given name is the top choice (or is NULL), zero is returned.  
* If the name does not appear in the list at all, NumChoices is returned.
****************************************************************************/
int Profile::GetOrdinal( const char *symbname ) const
    {
    if( symbname == NULL ) return 0;
    for( int i = 0; i < NumChoices(); i++ )
	{
        if( Match( symbname, GetSymbol(i)->Name() ) ) return i;
	}
    return NumChoices();
    }

/***************************************************************************
* Look up the index of the choice that matches the given symbol pointer.
* If the given symbol is the top choice (or is NULL), zero is returned.  
* If the symbol does not appear in the list at all, NumChoices is returned.
****************************************************************************/
int Profile::GetOrdinal( const Symbol *sym ) const
    {
    if( sym == NULL ) return 0;
    for( int i = 0; i < NumChoices(); i++ )
	{
        if( sym == GetSymbol(i) ) return i;
	}
    return NumChoices();
    }

ostream &operator<<( ostream &out, const Profile &prof ) 
    {
    out << "Profile" << endl;
    for( int i = 0; i < prof.NumChoices(); i++ )
	{
        Symbol *symbol = prof.GetSymbol(i);
        float   prob   = prof.GetProb  (i);
        out << form( "  %2d: %7.5f ", i+1, prob )
            << form( "\"%s\", from file %s", symbol->Name(), symbol->Qual() )
            << endl;
	}
    out << endl;
    return out;
    }

istream &operator>>( istream &in, Profile & ) 
    {
    assert(0);  // Not implemented.
    return in;
    }

