/***************************************************************************
* eps2fdf.C                                                                *
*                                                                          *
*  This program reads postscript files (specifically, those generated by   *
*  Adobe Illustrator) and generates a "font definition file" (fdf), which  *
*  is read by the "ReadDirectory" method of the "Font" class.              *
*                                                                          *
*  The postscript file must follow a specific format (meaning that the     *
*  objects must be in a particular order):                                 *
*                                                                          *
*      (#<name>)                                                           *
*      ...                                                                 *
*      [1 0 0 1 <x> <y> 0 Tp]                                              *
*      ...                                                                 *
*      (E) Tx                                                              *
*      ...                                                                 *
*      1 0 0 1 <x> <y> 0 Tp                                                *
*      ...                                                                 *
*      (S) Tx                                                              *
*      ...                                                                 *
*      <x> <y> m                                                           *
*      <x> <y> l                                                           *
*      <x> <y> l                                                           *
*      <x> <y> l                                                           *
*      <x> <y> l                                                           *
*      ...                                                                 *
*      <x> <y> m                                                           *
*      <x> <y> l                                                           *
*      <x1> <y1> <x2> <y2> <x3> <y3> c                                     *
*      ...                                                                 *
*      *u                                                                  *
*      ...                                                                 *
*      <x> <y> m                                                           *
*      <x> <y> l                                                           *
*      <x1> <y1> <x2> <y2> <x3> <y3> c                                     *
*      ...                                                                 *
*      (#end)                                                              *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       01/04/2000 arvo   Added support for triangulating outlines.        *
*       05/31/1998 arvo   Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, 2000 James Arvo                                      *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include <util/Char.h>
#include <rec/Font.h>

static char  buff[1024];
static char  name[64];
static char  *line = NULL;
static FILE  *file_in;
       FILE  *file_out;  // Not static because Triangulate.C needs it
static int   line_num = 0;
static float xmin, xmax;
static float ymin, ymax;
static float scale;
static float x_orig, y_orig;
static float exp_x, exp_y;
static float sub_x, sub_y;
static float x_coord, y_coord;
static int   quit = 0;
static int   verbose = 0;
static int   triangulate = 0;
static int   add_to_skeleton = 0;
static int   add_to_outline  = 0;
static Multiline skeleton;
static Multiline outline;
static Vec2  start;

extern void Open( const char *thing );
extern void Close();

extern void GenerateTriangles( const Multiline &skeleton, const Multiline &outline );


static inline void Transform( float &x, float &y )
    {
    x = ( x - x_orig ) / scale;
    y = ( y - y_orig ) / scale;
    }

static inline int notWhite( char c ) 
    { 
    return c != NewLine && c != NullChar; 
    }

static int Error( int flag )
    {
    char *msg = NULL;
    cerr << "ERROR (eps2fdf): Bad format while reading [" << name << "]." << endl;
    switch( flag )
        {
        case 100: msg = "Bad coors of exponent or subscript"    ; break;
        case 101: msg = "Reading first coord of bounding box"   ; break;
        case 102: msg = "Reading coord of bounding box"         ; break;
        case 103: msg = "Reading Move command"                  ; break;
        case 104: msg = "Reading Line command"                  ; break;
        case 105: msg = "Reading Curve command"                 ; break;
        case 106: msg = "Block ended prematurely"               ; break;
        case 107: msg = "Expected Line command in bounding box" ; break;
        case 108: msg = "File ended prematurely"                ; break;
        case 109: msg = "Reading Circ command"                  ; break;
        }
    cerr << msg << " at line " << line_num << endl;
    exit(1);
    return flag;
    }

static int Ends( char ch )
    {
    if( line == NULL || *line == NullChar ) return 0;
    char *s = line;
    while( notWhite( *(s + 1) ) != NullChar ) s++;
    return ToUpper( *s ) == ToUpper( ch );
    }

static int Begins( const char *pat )
    {
    if( line == NULL || *line == NullChar ) return 0;
    if( pat  == NULL || *pat  == NullChar ) return 0;
    char *s = line;
    while( ToUpper( *s ) == ToUpper( *pat ) )
        {
        if( *pat++ == NullChar ) return 1;
        if( *s++   == NullChar ) return 0;
        }
    if( *pat == NullChar ) return 1;
    return 0;
    }

static void ReadLine()
    {
    line = fgets( buff, 1024, file_in  );
    if( line != NULL ) line_num++;
    }

static void Start()
    {
    char buff[128];
    int k = 0;
    int literal = 0;
    for( int i = 2; line[i] != NullChar; i++ )
        {
        char ch = line[i];
        if( literal ) 
            {
            buff[k++] = ch;
            literal = 0;
            }
        else if( ch == Backslash ) 
            literal = 1;
        else if( ch == Rparen || ch == NewLine ) 
            break;
        else
            buff[k++] = ch;
        }
    buff[k] = NullChar;
    if( Match( buff, "exit" ) )
        {
        quit = 1;
        return;
        }
    fprintf( file_out, "\"%s\" {\n", buff );
    strcpy( name, buff );
    if( verbose ) printf( "      %s\n", name );
    sub_x = sub_y = 0.0;
    exp_x = exp_y = 0.0;
    }

void Open( const char *thing )
    {
    if( triangulate )
        {
        add_to_skeleton = 0;
        add_to_outline  = 0;
        if( Match( thing, "stroke"  ) ) { add_to_skeleton = 1; skeleton.Clear(); }
        if( Match( thing, "outline" ) ) { add_to_outline  = 1; outline.Clear();  }
        }
    fprintf( file_out, "  %s {\n", thing );
    }

void Close()
    {
    fprintf( file_out, "    }\n" );
    }

static void End()
    {
    if( sub_x != 0.0 )
        {
        Transform( sub_x, sub_y );
        fprintf( file_out, "  sub_pos %8.5f %8.5f\n", sub_x, sub_y );    
        }
    if( exp_x != 0.0 )
        {
        Transform( exp_x, exp_y );
        fprintf( file_out, "  exp_pos %8.5f %8.5f\n", exp_x, exp_y );    
        }
    if( triangulate ) 
        {
        GenerateTriangles( skeleton, outline );
        skeleton.Clear();
        outline.Clear();
        }
    fprintf( file_out, "  }\n\n" );
    }

static int Eof()
    {
    return feof( file_in  );
    }

static void SetCoord()
    {
    float x, y;
    if( sscanf( line, "1 0 0 1 %f %f", &x, &y ) != 2 ) Error(100);
    x_coord = x;
    y_coord = y;
    }

static void ExpPos()
    {
    exp_x = x_coord;
    exp_y = y_coord;
    }

static void SubPos()
    {
    sub_x = x_coord;
    sub_y = y_coord;
    }

static void NewBox()
    {
    float x, y;
    if( sscanf( line, "%f %f", &x, &y ) != 2 ) Error(101);
    xmin = x;
    xmax = x;
    ymin = y;
    ymax = y;
    }

static void Box()
    {
    float x, y;
    if( sscanf( line, "%f %f", &x, &y ) != 2 ) Error(102);
    if( x < xmin ) xmin = x;
    if( x > xmax ) xmax = x;
    if( y < ymin ) ymin = y;
    if( y > ymax ) ymax = y;
    }

static void EndBox()
    {
    Box();
    scale  = ymax - ymin;
    x_orig = xmin;
    y_orig = ymin;
    float width = ( xmax - xmin ) / scale;
    fprintf( file_out, "  width %f\n", width );
    }

static void Move()
    {
    float x, y;
    if( sscanf( line, "%f %f", &x, &y ) != 2 ) Error(103);
    Transform( x, y );
    start = Vec2( x, y );
    fprintf( file_out, "    m %8.5f %8.5f\n", x, y );
    if( add_to_skeleton ) { skeleton << Multiline::Jump; skeleton << Vec2( x, y ); }
    if( add_to_outline  ) { outline  << Multiline::Jump; outline  << Vec2( x, y ); }
    }

static void Line()
    {
    float x, y;
    if( sscanf( line, "%f %f", &x, &y ) != 2 ) Error(104);
    Transform( x, y );
    start = Vec2( x, y );
    fprintf( file_out, "    l %8.5f %8.5f\n", x, y );
    if( add_to_skeleton ) { skeleton << Vec2( x, y ); }
    if( add_to_outline  ) { outline  << Vec2( x, y ); }
    }

static void Curv()
    {
    float bx, by, cx, cy, dx, dy;
    if( sscanf( line, "%f %f %f %f %f %f", 
                &bx, &by, &cx, &cy, &dx, &dy ) != 6 ) Error(105);
    Transform( bx, by );
    Transform( cx, cy );
    Transform( dx, dy );
    if( triangulate )
        {
        // Do not store any curves -- rather, break them all down into polylines
        // since that is what will be triangulated anyway.
        Multiline C( Subdivide( start, Vec2(bx,by), Vec2(cx,cy), Vec2(dx,dy) ) );
        for( Pnode *v = C.First()->Next(); v != NULL; v = v->Next() )
            {
            fprintf( file_out, "    l %8.5f %8.5f\n", v->X(), v->Y() );
            if( add_to_skeleton ) { skeleton << *v; }
            if( add_to_outline  ) { outline  << *v; }
            }
        start = Vec2( dx, dy );
        }
    else
        {
        fprintf( file_out, "    c %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n", 
            bx, by, cx, cy, dx, dy );
        }
    }

static void Circ()
    {
    float bx, by, cx, cy;
    if( sscanf( line, "%f %f %f %f", &bx, &by, &cx, &cy ) != 4 ) Error(109);
    Transform( bx, by );
    Transform( cx, cy );
    fprintf( file_out, "    l %8.5f %8.5f\n", cx, cy );
    if (triangulate)
        {
        if( add_to_skeleton ) { skeleton << Vec2( cx, cy ); }
        if( add_to_outline  ) { outline  << Vec2( cx, cy ); }
        }
    }


static int convert( FILE *fp_in, FILE *fp_out )
    {
    int state = 0;
    file_in   = fp_in;
    file_out  = fp_out;
    for(;;)
        {
        ReadLine();
        switch( state )
            {
            case 0:
                if( Begins( "(#" ) ) { Start(); state++; }
                else if( Eof() ) return 0;
                if( quit ) return 0;
                break;

            case 1:
                if( Begins( "1 0 0 1" ) ) SetCoord(); else
                if( Begins( "(E)"     ) ) ExpPos  (); else
                if( Begins( "(S)"     ) ) SubPos  (); else
                if( Ends  ( 'M'       ) ) { NewBox(); state++; } else
                if( Eof() || Begins( "(#" ) ) return Error(106);
                break;

            case 2:
            case 3:
            case 4:
                if( Ends( 'L' ) ) { Box(); state++; }
                else return Error(107);
                break;

            case 5:
                if( Ends( 'L' ) ) { EndBox(); state++; }
                else return Error(107);
                break;

            case 6:
                if( Ends( 'M' ) ) { Open( "stroke" ); Move(); state++; }
                else if( Eof() || Begins( "(#" ) ) return Error(106);
                break;

            case 7:
                if( Ends  ( 'M'  ) ) Move(); else
                if( Ends  ( 'L'  ) ) Line(); else
                if( Ends  ( 'C'  ) ) Curv(); else
                if( Ends  ( 'Y'  ) ) Circ(); else
                if( Ends  ( 'V'  ) ) Circ(); else
                if( Begins( "*u" ) ) { Close(); Open( "outline" ); state++; } else
                if( Eof() ) return Error(108) ; else
                if( Begins( "(#" ) ) return 0;
                break;

            case 8:
                if( Ends( 'M' ) ) { Move(); state++; }
                else if( Eof() || Begins( "(#" ) ) return Error(106);
                break;

            case 9:
                if( Ends( 'M' ) ) Move(); else
                if( Ends( 'L' ) ) Line(); else
                if( Ends( 'C' ) ) Curv(); else
                if( Ends( 'Y' ) ) Circ(); else
                if( Ends( 'V' ) ) Circ(); else
                if( Begins( "(#" ) ) { Close(); End(); state = 0; } else
                if( Eof() ) return Error(108);
                break;
            }
        }
    }


void main( int argc, char *argv[] )
    {
    char out_dir[256];
    out_dir[0] = NullChar;
    name   [0] = NullChar;
    if( argc < 2 ) 
        {
        cerr << "Usage: eps2fdf [-t | -o <dir>] <file1> <file2> ... <fileN>" << endl;
        exit(0);
        }

    // Loop through all ".eps" files listed on the command line and generate
    // corresponging ".fdf" files.  Also look for "-o" flags that indicate where
    // to put the ".fdf" files.  By default, they are written to the directory
    // in which the corresponding ".eps" file is found.
    //
    for( int i = 1; i < argc; i++ )
        {
        const char *arg = argv[i];

        if( Match( arg, "-t" ) )
            {
            triangulate = 1;
            continue;
            }

        if( Match( arg, "-o" ) )
            {
            i++;
            if( i < argc ) strcpy( out_dir, argv[i] );
            continue;
            }

        char fname_in [128];
        char fname_out[128];
        strcpy( fname_in, arg );
        if( fname_in[0] == Percent ) continue;  // Skip auxiliary Illustrator files.
        FILE *fp_in = fopen( fname_in, "r" );
        if( fp_in == NULL )
            {
            strcat( fname_in, ".eps" );
            fp_in = fopen( fname_in, "r" );
            }
        if( fp_in == NULL )
            {
            cerr << "WARNING (eps2fdf): could not open file " << arg << endl;
            continue;
            }

        // Append the ".fdf" suffix, after stripping the ".eps" suffix (if necessary).
        if( !stripSuffix( fname_in, ".eps", fname_out ) )
            strcpy( fname_out, fname_in );
        strcat( fname_out, ".fdf" );

        // Change the qualifying path if one was specified on the command line.
        if( !nullString( out_dir ) )
            {
            char temp[128];
            getFile( fname_out, temp );
            sprintf( fname_out, "%s/%s", out_dir, temp );
            }

        FILE *fp_out = fopen( fname_out, "w" );
        if( fp_out == NULL )
            {
            cerr << "ERROR (eps2fdf): could not open file " << fname_out << endl;
            continue;
            }

        int len = strlen( fname_in );
        printf( "   %s ", fname_in );
        for( int j = 0; j < 26 - len; j++ ) printf( " " );
        printf( " ==>  %s\n", fname_out );
        convert( fp_in, fp_out );
        fclose( fp_in  );
        fclose( fp_out );
        }
    }

