%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% GetTeX.Txl
%	- One pass translation from .dat files to TeX.
%
% Copyright (C) 2000,2002 Richard Zanibbi
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%
% You may contact the authors by email at zanibbi@cs.queensu.ca,
% or by mail at: Richard Zanibbi, Dept. Computing Science,
% Queen's University, Kingston, Ontario, CANADA, K7L 3N6
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revision History
% v 1.0.0 February 14, 2001
% v 1.0.1 Mar 1, 2001
%	- made it so that the intermediate
%	  information is written out.
%	   (may want to make this a command line arg?)
% v 1.1.0 January 21, 2002
%	- added command line arguments to allow more
%	  flexible use of DRACULAE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Load BST Grammar and Symbol Class definitions.
include "Grammars/BST.Grammar"
include "Grammars/Symbol_Classes.Grammar"

% Rules needed to create TeX output.
include "Rules/Layout_Analysis.Rules"
include "Rules/Token_Lexing.Rules"
include "Rules/Relation_Lexing.Rules"
include "Rules/TeX_Symbol_List.Rules"
include "Rules/TeX_Output.Rules"


% File output and origin moving functions.
include "Rules/Files.Rules"
include "Rules/MoveOrigin.Rules"

% Code to handle command line arguments
include "Rules/CommandLineArgs.Rules"

function main
	replace [program]
	    P[program]		

	import TXLargs[repeat stringlit]
	
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% Default global settings (configurable from command line)
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	construct IntermediateDir [stringlit]
		"./"
	
	construct IntermediateFileName[stringlit]
		"DRACULAE.bst"
	
	% By default, do not write out intermediate data.
	construct GenerateOutput [stringlit]
		"False"

	construct ThresholdRatio[number]
		_[+ 1][/ 6]

	construct CentroidRatio[number]
		_[+ 1][/ 4]

	construct OriginLocation[stringlit]
		"Top Left"

	construct Version [stringlit]
		"0.2"

	construct HelpRequested[stringlit]
		"False"

	export IntermediateDir
	export IntermediateFileName
	export GenerateOutput
	export ThresholdRatio
	export CentroidRatio
	export OriginLocation
	export Version
	export HelpRequested

	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% Process command line arguments
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	construct Args[repeat stringlit]
		TXLargs[processArguments]
		#ifdef DEBUG
		[showArgs]  %for debugging.
		#endif

	% Re-import these values; they may have changed.
	import IntermediateFileName
	import OriginLocation
	import HelpRequested

	where 
		HelpRequested[= "False"]

	by
	    P[writeIntermediateFileHeader IntermediateFileName] %file header.
		 [processFFESexpression] % for an FFES expression.
		 [processExpressionList] % for a list of expressions.
end function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Functions controlling application of analysis.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function processExpression
	replace [baselineStructureTree]
		BST[baselineStructureTree]

	import IntermediateFileName[stringlit]

	by
		BST[analyzeLayout]   % Analyze symbol layout (build BST)
		 [appendToFile IntermediateFileName "I. Baseline Structure Tree"]
		 
		 [LexTokens]
		 [appendToFile IntermediateFileName "II. Token Lexed BST"]
		 
	   	 [LexRelations]
		 [appendToFile IntermediateFileName "III. Relation Lexed BST"]

	     [createTeX]
end function

function processFFESexpression
	% For a single expression in "FFES" .dat format.
	replace [program]
		H[ffesHeader] BST[baselineStructureTree]

	by
		BST[processExpression]
end function

function processUWExpression
	replace [UW_III_Expression]
		Label[expressionLabel] BST[baselineStructureTree]
	by
		'\bigskip '% Label '\begin{displaymath} BST[processExpression]
		'\end{displaymath}
end function

% Currently this is unused; eventually it would be nice to be able
% to do.
%rule unquoteSymbols
%	replace $ [symbol]
%		S[stringlit]

%	construct SID[symbol_id_attributes]
%		_[parse S]

%	by
%		"SYMBOL" SID
%end rule

function processExpressionList
	% For a list of expressions (e.g. generated from UW_III)
	replace * [repeat UW_III_Expression]
	 	expression[UW_III_Expression] Rest[repeat UW_III_Expression]

	import IntermediateFileName[stringlit]

	deconstruct expression
		Label[expressionLabel] BST[baselineStructureTree]

	construct ExpressionLabel[stringlit]
		_[quote Label]
		
	construct Space[stringlit]
		""
		
	construct FileLabel[stringlit]
		Space[appendToFile IntermediateFileName ExpressionLabel]
		
	by
		expression[processUWExpression] Rest[processExpressionList]
end function





